from anndata import AnnData
import scanpy as sc
import numpy as np
import pandas as pd
from scipy import sparse
from typing import Iterable, Union, Optional
from .._settings import add_reference


### Refer to: https://github.com/theislab/scanpy/blob/5533b644e796379fd146bf8e659fd49f92f718cd/scanpy/_compat.py
try:
    from typing import Literal
except ImportError:
    try:
        from typing_extensions import Literal
    except ImportError:

        class LiteralMeta(type):
            def __getitem__(cls, values):
                if not isinstance(values, tuple):
                    values = (values,)
                return type('Literal_', (Literal,), dict(__args__=values))

        class Literal(metaclass=LiteralMeta):
            pass

### Refer to Scanpy       
def _select_top_n(scores, n_top):
    reference_indices = np.arange(scores.shape[0], dtype=int)
    partition = np.argpartition(scores, -n_top)[-n_top:]
    partial_indices = np.argsort(scores[partition])[::-1]
    global_indices = reference_indices[partition][partial_indices]
    return global_indices


### Import from Scanpy
from scipy.sparse import issparse

### Import from Scanpy
def select_groups(adata, groups_order_subset='all', key='groups'):
    """Get subset of groups in adata.obs[key]."""
    groups_order = adata.obs[key].cat.categories
    if key + '_masks' in adata.uns:
        groups_masks = adata.uns[key + '_masks']
    else:
        groups_masks = np.zeros(
            (len(adata.obs[key].cat.categories), adata.obs[key].values.size), dtype=bool
        )
        for iname, name in enumerate(adata.obs[key].cat.categories):
            # if the name is not found, fallback to index retrieval
            if adata.obs[key].cat.categories[iname] in adata.obs[key].values:
                mask = adata.obs[key].cat.categories[iname] == adata.obs[key].values
            else:
                mask = str(iname) == adata.obs[key].values
            groups_masks[iname] = mask
    groups_ids = list(range(len(groups_order)))
    if groups_order_subset != 'all':
        groups_ids = []
        for name in groups_order_subset:
            groups_ids.append(
                np.where(adata.obs[key].cat.categories.values == name)[0][0]
            )
        if len(groups_ids) == 0:
            # fallback to index retrieval
            groups_ids = np.where(
                np.in1d(
                    np.arange(len(adata.obs[key].cat.categories)).astype(str),
                    np.array(groups_order_subset),
                )
            )[0]
        if len(groups_ids) == 0:
            logg.debug(
                f'{np.array(groups_order_subset)} invalid! specify valid '
                f'groups_order (or indices) from {adata.obs[key].cat.categories}',
            )
            from sys import exit

            exit(0)
        groups_masks = groups_masks[groups_ids]
        groups_order_subset = adata.obs[key].cat.categories[groups_ids].values
    else:
        groups_order_subset = groups_order.values
    return groups_order_subset, groups_masks


### Import from Scanpy   
import numpy as np
from scipy import sparse
import numba




def _get_mean_var(X, *, axis=0):
    if sparse.issparse(X):
        mean, var = sparse_mean_variance_axis(X, axis=axis)
    else:
        mean = np.mean(X, axis=axis, dtype=np.float64)
        mean_sq = np.multiply(X, X).mean(axis=axis, dtype=np.float64)
        var = mean_sq - mean ** 2
    # enforce R convention (unbiased estimator) for variance
    var *= X.shape[axis] / (X.shape[axis] - 1)
    return mean, var


def sparse_mean_variance_axis(mtx: sparse.spmatrix, axis: int):
    """
    This code and internal functions are based on sklearns
    `sparsefuncs.mean_variance_axis`.
    Modifications:
    * allow deciding on the output type, which can increase accuracy when calculating the mean and variance of 32bit floats.
    * This doesn't currently implement support for null values, but could.
    * Uses numba not cython
    """
    assert axis in (0, 1)
    if isinstance(mtx, sparse.csr_matrix):
        ax_minor = 1
        shape = mtx.shape
    elif isinstance(mtx, sparse.csc_matrix):
        ax_minor = 0
        shape = mtx.shape[::-1]
    else:
        raise ValueError("This function only works on sparse csr and csc matrices")
    if axis == ax_minor:
        return sparse_mean_var_major_axis(
            mtx.data, mtx.indices, mtx.indptr, *shape, np.float64
        )
    else:
        return sparse_mean_var_minor_axis(mtx.data, mtx.indices, *shape, np.float64)


@numba.njit(cache=True)
def sparse_mean_var_minor_axis(data, indices, major_len, minor_len, dtype):
    """
    Computes mean and variance for a sparse matrix for the minor axis.
    Given arrays for a csr matrix, returns the means and variances for each
    column back.
    """
    non_zero = indices.shape[0]

    means = np.zeros(minor_len, dtype=dtype)
    variances = np.zeros_like(means, dtype=dtype)

    counts = np.zeros(minor_len, dtype=np.int64)

    for i in range(non_zero):
        col_ind = indices[i]
        means[col_ind] += data[i]

    for i in range(minor_len):
        means[i] /= major_len

    for i in range(non_zero):
        col_ind = indices[i]
        diff = data[i] - means[col_ind]
        variances[col_ind] += diff * diff
        counts[col_ind] += 1

    for i in range(minor_len):
        variances[i] += (major_len - counts[i]) * means[i] ** 2
        variances[i] /= major_len

    return means, variances


@numba.njit(cache=True)
def sparse_mean_var_major_axis(data, indices, indptr, major_len, minor_len, dtype):
    """
    Computes mean and variance for a sparse array for the major axis.
    Given arrays for a csr matrix, returns the means and variances for each
    row back.
    """
    means = np.zeros(major_len, dtype=dtype)
    variances = np.zeros_like(means, dtype=dtype)

    for i in range(major_len):
        startptr = indptr[i]
        endptr = indptr[i + 1]
        counts = endptr - startptr

        for j in range(startptr, endptr):
            means[i] += data[j]
        means[i] /= minor_len

        for j in range(startptr, endptr):
            diff = data[j] - means[i]
            variances[i] += diff * diff

        variances[i] += (minor_len - counts) * means[i] ** 2
        variances[i] /= minor_len

    return means, variances  

### Import from Scanpy
class _RankGenes:
    def __init__(
        self,
        adata,
        groups,
        groupby,
        reference='rest',
        use_raw=True,
        layer=None,
        comp_pts=False,
    ):

        if 'log1p' in adata.uns_keys() and adata.uns['log1p']['base'] is not None:
            self.expm1_func = lambda x: np.expm1(x * np.log(adata.uns['log1p']['base']))
        else:
            self.expm1_func = np.expm1

        self.groups_order, self.groups_masks = select_groups(
            adata, groups, groupby
        )

        # Singlet groups cause division by zero errors
        invalid_groups_selected = set(self.groups_order) & set(
            adata.obs[groupby].value_counts().loc[lambda x: x < 2].index
        )

        if len(invalid_groups_selected) > 0:
            raise ValueError(
                "Could not calculate statistics for groups {} since they only "
                "contain one sample.".format(', '.join(invalid_groups_selected))
            )

        adata_comp = adata
        if layer is not None:
            if use_raw:
                raise ValueError("Cannot specify `layer` and have `use_raw=True`.")
            X = adata_comp.layers[layer]
        else:
            if use_raw and adata.raw is not None:
                adata_comp = adata.raw
            X = adata_comp.X

        # for correct getnnz calculation
        if issparse(X):
            X.eliminate_zeros()

        self.X = X
        self.var_names = adata_comp.var_names

        self.ireference = None
        if reference != 'rest':
            self.ireference = np.where(self.groups_order == reference)[0][0]

        self.means = None
        self.vars = None

        self.means_rest = None
        self.vars_rest = None

        self.comp_pts = comp_pts
        self.pts = None
        self.pts_rest = None

        self.stats = None

        # for logreg only
        self.grouping_mask = adata.obs[groupby].isin(self.groups_order)
        self.grouping = adata.obs.loc[self.grouping_mask, groupby]

    def _basic_stats(self):
        n_genes = self.X.shape[1]
        n_groups = self.groups_masks.shape[0]

        self.means = np.zeros((n_groups, n_genes))
        self.vars = np.zeros((n_groups, n_genes))
        self.pts = np.zeros((n_groups, n_genes)) if self.comp_pts else None

        if self.ireference is None:
            self.means_rest = np.zeros((n_groups, n_genes))
            self.vars_rest = np.zeros((n_groups, n_genes))
            self.pts_rest = np.zeros((n_groups, n_genes)) if self.comp_pts else None
        else:
            mask_rest = self.groups_masks[self.ireference]
            X_rest = self.X[mask_rest]
            self.means[self.ireference], self.vars[self.ireference] = _get_mean_var(
                X_rest
            )
            # deleting the next line causes a memory leak for some reason
            del X_rest

        if issparse(self.X):
            get_nonzeros = lambda X: X.getnnz(axis=0)
        else:
            get_nonzeros = lambda X: np.count_nonzero(X, axis=0)

        for imask, mask in enumerate(self.groups_masks):
            X_mask = self.X[mask]

            if self.comp_pts:
                self.pts[imask] = get_nonzeros(X_mask) / X_mask.shape[0]

            if self.ireference is not None and imask == self.ireference:
                continue

            self.means[imask], self.vars[imask] = _get_mean_var(X_mask)

            if self.ireference is None:
                mask_rest = ~mask
                X_rest = self.X[mask_rest]
                self.means_rest[imask], self.vars_rest[imask] = _get_mean_var(X_rest)
                # this can be costly for sparse data
                if self.comp_pts:
                    self.pts_rest[imask] = get_nonzeros(X_rest) / X_rest.shape[0]
                # deleting the next line causes a memory leak for some reason
                del X_rest

        
def cosg(
    adata,
    groupby='CellTypes',
    groups: Union[Literal['all'], Iterable[str]] = 'all',

    mu=1,
    remove_lowly_expressed:bool=False,
    expressed_pct:Optional[float] = 0.1,

    n_genes_user:int =50,
    key_added: Optional[str] = None,
    calculate_logfoldchanges: bool = True,
    use_raw: bool = True,
    layer: Optional[str] = None,
    reference: str = 'rest',    

    copy:bool=False
):
    """
    Marker gene identification for single-cell sequencing data using COSG.
    
    Arguments:
        adata: Annotated data matrix. Note: input parameters are similar to the parameters used for scanpy's rank_genes_groups() function.
        groupby: The key of the cell groups in .obs, the default value is set to 'CellTypes'.
        groups: Subset of cell groups, e.g. ['g1', 'g2', 'g3'], to which comparison shall be restricted. The default value is 'all', and all groups will be compared.
        mu: The penalty restricting marker genes expressing in non-target cell groups. Larger value represents more strict restrictions. mu should be >= 0, and by default, mu = 1.
        remove_lowly_expressed: If True, genes that express a percentage of target cells smaller than a specific value (expressed_pct) are not considered as marker genes for the target cells. The default value is False.
        expressed_pct: When remove_lowly_expressed is set to True, genes that express a percentage of target cells smaller than a specific value (expressed_pct) are not considered as marker genes for the target cells. The default value for expressed_pct is 0.1 (10%).
        n_genes_user: The number of genes that appear in the returned tables. The default value is 50.
        key_added: The key in adata.uns information is saved to.
        calculate_logfoldchanges: Calculate logfoldchanges.
        use_raw: Use raw attribute of adata if present.
        layer: Key from adata.layers whose value will be used to perform tests on.
        reference: If 'rest', compare each group to the union of the rest of the group. If a group identifier, compare with respect to this group.

    Returns:
        names : structured np.ndarray (.uns['rank_genes_groups'])
                Structured array to be indexed by group id storing the gene names. Ordered according to scores.
        scores : structured np.ndarray (.uns['rank_genes_groups'])
                Structured array to be indexed by group id storing COSG scores for each gene for each group. Ordered according to scores.

    Notes:
        Contact: Min Dai, daimin@zju.edu.cn

    Examples:
        >>> import cosg as cosg
        >>> import scanpy as sc
        >>> adata = sc.datasets.pbmc68k_reduced()
        >>> cosg.cosg(adata, key_added='cosg', groupby='bulk_labels')
        >>> sc.pl.rank_genes_groups(adata, key='cosg')

    """
    
    adata = adata.copy() if copy else adata
        
    if layer is not None:
        if use_raw:
            raise ValueError("Cannot specify `layer` and have `use_raw=True`.")
        cellxgene = adata.layers[layer]
    else:
        if use_raw and adata.raw is not None:
             cellxgene = adata.raw.X
        cellxgene = adata.X
    
    
    ### Refer to scanpy's framework
    # https://github.com/theislab/scanpy/blob/5533b644e796379fd146bf8e659fd49f92f718cd/scanpy/tools/_rank_genes_groups.py#L559
    if key_added is None:
        key_added = 'rank_genes_groups'
    adata.uns[key_added] = {}
    adata.uns[key_added]['params'] = dict(
    groupby=groupby,
    reference=reference,
    groups=groups,
    method='cosg',
    use_raw=use_raw,
    layer=layer,
    )
    
    ### Refer to: https://github.com/theislab/scanpy/blob/5533b644e796379fd146bf8e659fd49f92f718cd/scanpy/tools/_rank_genes_groups.py#L543
    if groups == 'all':
        ### group lable for each cell
        group_info=adata.obs[groupby]
    elif isinstance(groups, (str, int)):
        raise ValueError('Specify a sequence of groups')
    else:
        cells_selected=adata.obs[groupby].isin(groups)
        cells_selected=cells_selected.values
        if sparse.issparse(cellxgene):
            cellxgene=cellxgene[cells_selected]
        else:
            cellxgene=cellxgene[cells_selected,:]
            
            

        ### group lable for each cell
        group_info=adata.obs[groupby].copy()
        group_info=group_info[cells_selected]
        

    
    groups_order=np.unique(group_info)
    n_cluster=len(groups_order)
    
    n_cell=cellxgene.shape[0]
    cluster_mat=np.zeros(shape=(n_cluster,n_cell))

    ### To further restrict expression in other clusters, can think about a better way, such as taking the cluster similarities into consideration
    order_i=0
    for group_i in groups_order:    
        idx_i=group_info==group_i 
        cluster_mat[order_i,:][idx_i]=1
        order_i=order_i+1
    
    if sparse.issparse(cellxgene):
        ### Convert to sparse matrix
        from scipy.sparse import csr_matrix
        cluster_mat=csr_matrix(cluster_mat)

        from sklearn.metrics.pairwise import cosine_similarity

        ### the dimension is: Gene x lambda
        cosine_sim=cosine_similarity(X=cellxgene.T,Y=cluster_mat,dense_output=False) 

        pos_nonzero=cosine_sim.nonzero()
        genexlambda=cosine_sim.multiply(cosine_sim)

        e_power2_sum=genexlambda.sum(axis=1)


        if mu==1:
            genexlambda[pos_nonzero]=genexlambda[pos_nonzero]/(np.repeat(e_power2_sum,genexlambda.shape[1],axis=1)[pos_nonzero])
                                                    
        else:
            genexlambda[pos_nonzero]=genexlambda[pos_nonzero]/((1-mu)*genexlambda[pos_nonzero]+
                                                     mu*(
                                                        np.repeat(e_power2_sum,genexlambda.shape[1],axis=1)[pos_nonzero])
                                                    )
        genexlambda[pos_nonzero]=np.multiply(genexlambda[pos_nonzero],cosine_sim[pos_nonzero])
    
    ### If the cellxgene is not a sparse matrix
    else:
         ## Not using sparse matrix
        from sklearn.metrics.pairwise import cosine_similarity    
        cosine_sim=cosine_similarity(X=cellxgene.T,Y=cluster_mat,dense_output=True) 

        pos_nonzero=cosine_sim!=0
        e_power2=np.multiply(cosine_sim,cosine_sim)
        e_power2_sum=np.sum(e_power2,axis=1)
        e_power2[pos_nonzero]=np.true_divide(e_power2[pos_nonzero],(1-mu)*e_power2[pos_nonzero]+mu*(np.dot(e_power2_sum.reshape(e_power2_sum.shape[0],1),np.repeat(1,e_power2.shape[1]).reshape(1,e_power2.shape[1]))[pos_nonzero]))
        e_power2[pos_nonzero]=np.multiply(e_power2[pos_nonzero],cosine_sim[pos_nonzero])
        genexlambda=e_power2

    ### Refer to scanpy
    rank_stats=None
    
    ### Whether to calculate logfoldchanges, because this is required in scanpy 1.8
    if calculate_logfoldchanges:
        ### Calculate basic stats
        ### Refer to Scanpy
        # for clarity, rename variable
        if groups == 'all':
            groups_order2 = 'all'
        elif isinstance(groups, (str, int)):
            raise ValueError('Specify a sequence of groups')
        else:
            groups_order2 = list(groups)
            if isinstance(groups_order2[0], int):
                groups_order2 = [str(n) for n in groups_order2]
            if reference != 'rest' and reference not in set(groups_order2):
                groups_order2 += [reference]
        if reference != 'rest' and reference not in adata.obs[groupby].cat.categories:
            cats = adata.obs[groupby].cat.categories.tolist()
            raise ValueError(
                f'reference = {reference} needs to be one of groupby = {cats}.'
            )
        pts=False
        anndata_obj = _RankGenes(adata, groups_order2, groupby, reference, use_raw, layer, pts)
        anndata_obj._basic_stats()
    
    
    ### Refer to Scanpy
    # for correct getnnz calculation
    ### get non-zeros for columns
    if sparse.issparse(cellxgene):
        cellxgene.eliminate_zeros()
    if sparse.issparse(cellxgene):
        get_nonzeros = lambda X: X.getnnz(axis=0)
    else:
        get_nonzeros = lambda X: np.count_nonzero(X, axis=0)
    
    order_i=0
    for group_i in groups_order:    
        idx_i=group_info==group_i 

        ### Convert to numpy array
        idx_i=idx_i.values

        ## Compare the most ideal case to the worst case
        if sparse.issparse(cellxgene):
            scores=genexlambda[:,order_i].toarray()[:,0]
        else:
            scores=genexlambda[:,order_i]

        
        ### Mask these genes expressed in less than 3 cells in the cluster of interest
        if remove_lowly_expressed:
            n_cells_expressed=get_nonzeros(cellxgene[idx_i])
            n_cells_i=np.sum(idx_i)
            scores[n_cells_expressed<n_cells_i*expressed_pct]= -1

        global_indices = _select_top_n(scores, n_genes_user)

        if rank_stats is None:
            idx = pd.MultiIndex.from_tuples([(group_i,'names')])
            rank_stats = pd.DataFrame(columns=idx)
        rank_stats[group_i, 'names'] = adata.var_names[global_indices]
        rank_stats[group_i, 'scores'] = scores[global_indices]
        
        if calculate_logfoldchanges:
            group_index=np.where(anndata_obj.groups_order==group_i)[0][0]
            if anndata_obj.means is not None:
                mean_group = anndata_obj.means[group_index]
                if anndata_obj.ireference is None:
                    mean_rest = anndata_obj.means_rest[group_index]
                else:
                    mean_rest = anndata_obj.means[anndata_obj.ireference]
                foldchanges = (anndata_obj.expm1_func(mean_group) + 1e-9) / (
                    anndata_obj.expm1_func(mean_rest) + 1e-9
                )  # add small value to remove 0's
                rank_stats[group_i, 'logfoldchanges'] = np.log2(
                    foldchanges[global_indices]
                )
    
        order_i=order_i+1
            
    ## Refer to scanpy
    if calculate_logfoldchanges:
        dtypes = {
        'names': 'O',
        'scores': 'float32',
        'logfoldchanges': 'float32',
        }
    else:
        dtypes = {
        'names': 'O',
        'scores': 'float32',
        }
    ###
    rank_stats.columns = rank_stats.columns.swaplevel()
    for col in rank_stats.columns.levels[0]:
        adata.uns[key_added][col]=rank_stats[col].to_records(
    index=False, column_dtypes=dtypes[col]
    )
    adata.uns[key_added]['pvals']=adata.uns['rank_genes_groups']['pvals']
    adata.uns[key_added]['pvals_adj']=adata.uns['rank_genes_groups']['pvals_adj']
        
    
        
    print('**finished identifying marker genes by COSG**')
    add_reference(adata,'COSG','marker gene identification with COSG')
        
    ### Return the result
    return adata if copy else None



import numpy as np
import pandas as pd
from tqdm import tqdm
import scipy.sparse as sp

from .Model import Model
from .utils import Transfer_pytorch_Data

import torch

import torch.nn.functional as F


def train_BINARY(adata, 
                pos_weight=10.0, 
                hidden_dims=[512, 30], 
                n_epochs=1000, 
                lr=0.001, 
                key_added='BINARY',
                gradient_clipping=5.,  
                weight_decay=0.0001, 
                verbose=True, 
                random_seed=0,  
                device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu'),
                ):
    """
    Train a binary representation
    
    Parameters:
    ----------
    adata : AnnData
        The annotated data matrix containing the samples to be embedded.
        
    pos_weight : float, optional (default=10.0)
        The positive weight parameter used in the BCEWithLogitsLoss.
        
    hidden_dims : list of int, optional (default=[512, 30])
        The dimensionality of the hidden layers in the neural network.
        
    n_epochs : int, optional (default=1000)
        Number of training epochs.
        
    lr : float, optional (default=0.001)
        The learning rate for the Adam optimizer.
        
    key_added : str, optional (default='BINARY')
        Key under which the binary representations are saved in the `obsm` slot of the `adata` object.
        
    gradient_clipping : float, optional (default=5.)
        Maximum allowed norm of the gradients for stabilization.
        
    weight_decay : float, optional (default=0.0001)
        Weight decay (L2 penalty) used in the Adam optimizer.
        
    verbose : bool, optional (default=True)
        Whether to print status messages during the process.
        
    random_seed : int, optional (default=0)
        Seed for random number generator for reproducibility.
        
    device : torch.device, optional
        The device (e.g., CPU or CUDA) on which the model and data should be loaded.
        
    Returns:
    -------
    AnnData
        The input `adata` object with the binary representations stored in `adata.obsm[key_added]`.
    """
    import torch.backends.cudnn as cudnn
    cudnn.deterministic = True
    cudnn.benchmark = True

    seed=random_seed
    import random
    random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)

    adata.X = sp.csr_matrix(adata.X)
    
    if 'highly_variable' in adata.var.columns:
        adata_Vars =  adata[:, adata.var['highly_variable']]
    else:
        adata_Vars = adata

    if verbose:
        print('Size of Input: ', adata_Vars.shape)
    if 'Spatial_Graph' not in adata.uns.keys():
        raise ValueError("Spatial_Graph is not existed! Run Construct_Spatial_Graph first!")

    data = Transfer_pytorch_Data(adata_Vars)

    model = Model(hidden_dims = [data.x.shape[1]] + hidden_dims).to(device)
    data = data.to(device)

    optimizer = torch.optim.Adam(model.parameters(), lr=lr, weight_decay=weight_decay)
    
    loss_list = []
    for epoch in tqdm(range(1, n_epochs+1)):
        model.train()
        optimizer.zero_grad()
        z, out = model(data.x, data.edge_index)
        criterion = torch.nn.BCEWithLogitsLoss(pos_weight=torch.tensor([pos_weight]).to(device))
        loss = criterion(out, data.x.float())
        loss_list.append(loss)
        loss.backward()
        torch.nn.utils.clip_grad_norm_(model.parameters(), gradient_clipping)
        optimizer.step()
    
    model.eval()
    z, out = model(data.x, data.edge_index)
    
    BINARY_rep = z.to('cpu').detach().numpy()
    adata.obsm[key_added] = BINARY_rep

    return adata


"""
Module to initialise a bioFAM model
"""

import numpy as np
import scipy as s
import scipy.stats as stats
from sys import path
import sklearn.decomposition
from sklearn.impute import SimpleImputer

from ..core.nodes import *

class initModel(object):
    def __init__(self, dim, data, lik, groups, seed):
        """
        PARAMETERS
        dim: dictionary with keyworded dimensionalities:
            N for the number of samples
            M for the number of views
            K for the number of factors or latent variables,
            D for the number of features per view
        data: list of length M with numpy arrays of dimensionality (N,Dm)
        lik: list of strings with length M
            likelihood for each view, choose from ('gaussian','poisson','bernoulli')
        """

        # Set seed
        np.random.seed(seed)

        # Set groups
        assert len(groups) == dim["N"], 'sample groups labels do not match number of samples'
        self.groups = groups
        self.groups_ix = np.unique(groups, return_inverse=True)[1] # convert groups into integers from 0 to n_groups


        # Set data
        self.data = data

        # Set likelihoods
        self.lik = lik

        # Set dimensionalities
        self.N = dim["N"]
        self.K = dim["K"]
        self.M = dim["M"]
        self.D = dim["D"]
        self.G = dim["G"]
        # self.G = len(np.unique(groups))

        self.nodes = {}

    def initZ(self, pmean=0., pvar=1., qmean="random", qvar=1., qE=None, qE2=None, Y=None, impute=False, weight_views = False):
        """Method to initialise the latent variables

        PARAMETERS
        ----------
        pmean: mean of the prior distribution
        pvar: (co)variance of the prior distribution
        qmean: initialisation of the mean of the variational distribution
            "random" for a random initialisation sampled from a standard normal distribution
            "pca" for an initialisation based on PCA
            "orthogonal" for a random initialisation with orthogonal factors
        qvar: initial value of the (co)variance of the variational distribution
        qE: initial value of the expectation of the variational distribution
        qE2: initial value of the second moment of the variational distribution
        Y: matrix to run PCA on (when qmean="pca")
        impute: logical value if to perform imputation before running PCA,
            this is only applicable when qmean="pca" and missing values (np.NaN) are present in the data
        weight_views: logical whether to weight the ELBO
        """

        ## Initialise prior distribution (P)

        # mean
        pmean = np.ones((self.N, self.K)) * pmean

        ## Initialise variational distribution (Q)

        # variance
        qvar = np.ones((self.N, self.K)) * qvar

        # mean
        if qmean is not None:
            if isinstance(qmean, str):

                # Random initialisation
                if qmean == "random":
                    qmean = stats.norm.rvs(loc=0, scale=1, size=(self.N, self.K))

                # Random initialisation with orthogonal factors
                elif qmean == "orthogonal":
                    pca = sklearn.decomposition.PCA(n_components=self.K, whiten=True)
                    pca.fit(stats.norm.rvs(loc=0, scale=1, size=(self.N, 9999)).T)
                    qmean = pca.components_.T

                # PCA initialisation
                elif qmean == "pca":
                    pca = sklearn.decomposition.PCA(n_components=self.K, whiten=True)
                    Ytmp = np.concatenate(Y, axis=1)

                    if impute == True:
                        if np.any(np.isnan(Ytmp)):
                            imp = SimpleImputer(missing_values=np.NaN, strategy="mean")
                            imp.fit(Ytmp)
                            Ytmp = imp.transform(Ytmp)

                    pca.fit(Ytmp)
                    qmean = pca.transform(Ytmp)

                # scale factor values from -1 to 1 (per factor)
                qmean = 2.*(qmean - np.min(qmean,axis=0))/np.ptp(qmean,axis=0)-1

            elif isinstance(qmean, np.ndarray):
                assert qmean.shape == (self.N, self.K), "Wrong shape for the expectation of the Q distribution of Z"

            elif isinstance(qmean, (int, float)):
                qmean = np.ones((self.N, self.K)) * qmean

            else:
                print("Wrong initialisation for Z")
                exit()

        # Initialise the node
        self.nodes["Z"] = Z_Node(
            dim=(self.N, self.K),
            pmean=pmean, pvar=pvar,
            qmean=qmean, qvar=qvar,
            qE=qE, qE2=qE2, weight_views = weight_views
        )

    def initZ_smooth(self, pmean=0., pvar=1., qmean="random", qvar=1., qE=None, qE2=None, Y=None, impute=False, weight_views = False):
        """Method to initialise the latent variables

        PARAMETERS
        ----------
        pmean: mean of the prior distribution
        pvar: (co)variance of the prior distribution
        qmean: initialisation of the mean of the variational distribution
            "random" for a random initialisation sampled from a standard normal distribution
            "pca" for an initialisation based on PCA
            "orthogonal" for a random initialisation with orthogonal factors
        qvar: initial value of the (co)variance of the variational distribution
        qE: initial value of the expectation of the variational distribution
        qE2: initial value of the second moment of the variational distribution
        Y: matrix to run PCA on (when qmean="pca")
        impute: logical value if to perform imputation before running PCA,
            this is only applicable when qmean="pca" and missing values (np.NaN) are present in the data
        weight_views: logical whether to weight the ELBO
        """

        ## Initialise prior distribution (P)

        # mean
        pmean = np.ones((self.N, self.K)) * pmean
        if isinstance(pvar, (int,float)):
            pvar = np.array([np.eye(self.N)*pvar for k in range(self.K)])

        ## Initialise variational distribution (Q)

        # variance
        qvar = np.array([np.eye(self.N)*qvar for k in range(self.K)])

        # mean
        if qmean is not None:
            if isinstance(qmean, str):

                # Random initialisation
                if qmean == "random":
                    qmean = stats.norm.rvs(loc=0, scale=1, size=(self.N, self.K))

                # Random initialisation with orthogonal factors
                elif qmean == "orthogonal":
                    pca = sklearn.decomposition.PCA(n_components=self.K, whiten=True)
                    pca.fit(stats.norm.rvs(loc=0, scale=1, size=(self.N, 9999)).T)
                    qmean = pca.components_.T

                # PCA initialisation
                elif qmean == "pca":
                    pca = sklearn.decomposition.PCA(n_components=self.K, whiten=True)
                    Ytmp = np.concatenate(Y, axis=1)

                    if impute == True:
                        if np.any(np.isnan(Ytmp)):
                            imp = SimpleImputer(missing_values=np.NaN, strategy="mean")
                            imp.fit(Ytmp)
                            Ytmp = imp.transform(Ytmp)

                    pca.fit(Ytmp)
                    qmean = pca.transform(Ytmp)

                # scale factor values from -1 to 1 (per factor)
                qmean = 2.*(qmean - np.min(qmean,axis=0))/np.ptp(qmean,axis=0)-1

            elif isinstance(qmean, np.ndarray):
                assert qmean.shape == (self.N, self.K), "Wrong shape for the expectation of the Q distribution of Z"

            elif isinstance(qmean, (int, float)):
                qmean = np.ones((self.N, self.K)) * qmean

            else:
                print("Wrong initialisation for Z")
                exit()

        # Initialise the node
        self.nodes["Z"] = Z_GP_Node_mv(
            dim=(self.N, self.K),
            pmean=pmean, pcov=pvar,
            qmean=qmean, qcov=qvar,
            qE=qE, 
            weight_views = weight_views
        )
        # self.nodes["Z"] = Z_GP_Node(
        #     dim=(self.N, self.K),
        #     pmean=pmean, pcov=pvar,
        #     qmean=qmean, qvar=qvar,
        #     qE=qE, 
        #     weight_views = weight_views
        # )

    def initU(self, pmean=0., pvar=1., qmean=0, qvar=1., qE=None, qE2=None, Y = None, impute=True, idx_inducing = None, weight_views = False): # prior hanp.diagonal covariance here, ls optimiation in Sigma node
        """Method to initialise the inducing points

         PARAMETERS
         ----------
        pmean: mean of the prior distribution
        pvar: (co)variance of the prior distribution
         qmean: initialisation of the mean of the variational distribution
             "random" for a random initialisation sampled from a standard normal distribution
             "pca" for an initialisation based on PCA
             "orthogonal" for a random initialisation with orthogonal factors
         qvar: initial value of the variance of the variational distribution
         qE: initial value of the expectation of the variational distribution
         qE2: initial value of the second moment of the variational distribution
         Y: matrix to run PCA on (when qmean="pca")
         impute: logical value if to perform imputation before running PCA,
             this is only applicable when qmean="pca" and missing values (np.NaN) are present in the data
         GP_factors: logical whether to use a Z node with GP prior or not
         """

        assert idx_inducing is not  None, "Stop: U nodes is used without inducing points"

        Nu = len(idx_inducing)

        ## Initialise prior distribution (P)
        # mean
        pmean = np.ones((Nu, self.K)) * pmean
        if isinstance(pvar, (int, float)):
            pvar = np.array([np.eye(Nu) * pvar for k in range(self.K)])

        ## Initialise variational distribution (Q)
        qmean = np.ones((Nu, self.K)) * qmean

        # variance
        qvar = np.array([np.eye(Nu) * qvar for k in range(self.K)])

        # Initialise the node
        self.nodes["U"] = U_GP_Node_mv(
            dim=(Nu, self.K),
            pmean=pmean, pcov=pvar,
            qmean=qmean, qcov=qvar,
            qE=qE, idx_inducing = idx_inducing,
            weight_views = weight_views
        )

    def initZgU(self, pmean = 0, pvar =1., qmean = "random", qvar = 1., qE = None, qE2 = None, Y = None,
                impute = True, idx_inducing = None, weight_views = False):
        assert idx_inducing is not None, "Stop: ZgU nodes is used without inducing points"

        ## Initialise prior distribution (P)

        # mean
        pmean = np.ones((self.N, self.K)) * pmean

        ## Initialise variational distribution (Q)

        # variance
        qvar = np.ones((self.N, self.K)) * qvar

        # mean
        if qmean is not None:
            if isinstance(qmean, str):

                # Random initialisation
                if qmean == "random":
                    qmean = stats.norm.rvs(loc=0, scale=1, size=(self.N, self.K))

                # Random initialisation with orthogonal factors
                elif qmean == "orthogonal":
                    pca = sklearn.decomposition.PCA(n_components=self.K, whiten=True)
                    pca.fit(stats.norm.rvs(loc=0, scale=1, size=(self.N, 9999)).T)
                    qmean = pca.components_.T

                # PCA initialisation
                elif qmean == "pca":
                    # whiten=True scales the principal components to match the prior N(0,1)
                    pca = sklearn.decomposition.PCA(n_components=self.K, whiten=True)
                    Ytmp = np.concatenate(Y, axis=1)

                    if impute == True:
                        if np.any(np.isnan(Ytmp)):
                            imp = SimpleImputer(missing_values=np.NaN, strategy="mean")
                            imp.fit(Ytmp)
                            Ytmp = imp.transform(Ytmp)

                    pca.fit(Ytmp)
                    qmean = pca.transform(Ytmp)

                # scale factor values from -1 to 1 (per factor)
                qmean = 2.*(qmean - np.min(qmean,axis=0))/np.ptp(qmean,axis=0)-1

            elif isinstance(qmean, np.ndarray):
                assert qmean.shape == (self.N, self.K), "Wrong shape for the expectation of the Q distribution of Z"

            elif isinstance(qmean, (int, float)):
                qmean = np.ones((self.N, self.K)) * qmean

            else:
                print("Wrong initialisation for Z")
                exit()

        self.nodes["Z"] = ZgU_node(
            dim=(self.N, self.K),
            pmean=pmean, pvar=pvar,
            qmean=qmean, qvar=qvar,
            qE=qE, qE2=qE2,
            idx_inducing = idx_inducing,
            weight_views = weight_views
        )

    def initSigma(self, sample_cov, start_opt = 20, n_grid = 10,
        opt_freq = 10, model_groups = False, use_gpytorch = False):

        self.Sigma = Sigma_Node(
            dim=(self.K,), 
            sample_cov=sample_cov, 
            groups=self.groups,
            start_opt=start_opt, 
            n_grid=n_grid, 
            opt_freq=opt_freq, 
            model_groups = model_groups
        )
        self.nodes["Sigma"] = self.Sigma

    def initSigma_sparse(self, sample_cov, start_opt = 20, n_grid = 10, idx_inducing = None, 
        opt_freq = 10, model_groups = False, use_gpytorch = False):

        self.Sigma = Sigma_Node_sparse(
            dim=(self.K,), 
            sample_cov=sample_cov, 
            groups=self.groups,
            start_opt=start_opt, 
            n_grid=n_grid, 
            idx_inducing=idx_inducing, 
            opt_freq=opt_freq, 
            model_groups = model_groups
        )
        self.nodes["Sigma"] = self.Sigma

    def initSigma_warping(self, sample_cov, start_opt = 20, n_grid = 10, 
        warping_freq = 20, warping_ref = 0, warping_open_begin = True, warping_open_end =True, warping_groups = None,
        opt_freq = 10, model_groups = False, use_gpytorch = False):

        self.Sigma = Sigma_Node_warping(
            dim=(self.K,), 
            sample_cov=sample_cov, 
            groups=self.groups,
            start_opt=start_opt, 
            n_grid=n_grid, 
            warping_freq=warping_freq, 
            warping_ref=warping_ref,
            warping_open_begin=warping_open_begin, 
            warping_open_end=warping_open_end,
            warping_groups=warping_groups,
            opt_freq=opt_freq, 
            model_groups = model_groups
        )
        self.nodes["Sigma"] = self.Sigma

    def initSZ(self, pmean_T0=0., pmean_T1=0., pvar_T0=1., pvar_T1=1., ptheta=1., qmean_T0=0., qmean_T1="random", qvar_T0=1.,
        qvar_T1=1., qtheta=1., qEZ_T0=None, qEZ_T1=None, qET=None, Y=None, impute=False, weight_views = False):
        """Method to initialise sparse factors with a spike and slab prior

        PARAMETERS
        ----------
        (...)
        """

        # TODO:
        # - add different ways to initialise the expectations of the posterior: random, orthogonal, PCA


        ## Initialise prior distribution (P)

        ## Initialise variational distribution (Q)
        if isinstance(qmean_T1, str):

            if qmean_T1 == "random":
                qmean_T1 = stats.norm.rvs(loc=0, scale=1, size=(self.N, self.K))
            elif qmean_T1 == "pca":
                pca = sklearn.decomposition.PCA(n_components=self.K, whiten=True)
                Ytmp = np.concatenate(Y, axis=1)

                if impute == True:
                    if np.any(np.isnan(Ytmp)):
                        imp = SimpleImputer(missing_values=np.NaN, strategy="mean")
                        imp.fit(Ytmp)
                        Ytmp = imp.transform(Ytmp)

                pca.fit(Ytmp)
                qmean_T1 = pca.transform(Ytmp)
            else:
                print("%s initialisation not implemented for Z" % qmean_T1)
                exit()

        elif isinstance(qmean_T1, np.ndarray):
            assert qmean_T1.shape == (self.N, self.K), "Wrong dimensionality"

        elif isinstance(qmean_T1, (int, float)):
            qmean_T1 *= np.ones((self.N, self.K))

        else:
            print("Wrong initialisation for Z")
            exit(1)

        self.nodes["Z"] = SZ_Node(
            dim=(self.N, self.K),

            ptheta=ptheta,
            pmean_T0=pmean_T0, pvar_T0=pvar_T0,
            pmean_T1=pmean_T1, pvar_T1=pvar_T1,

            qtheta=qtheta,
            qmean_T0=qmean_T0, qvar_T0=qvar_T0,
            qmean_T1=qmean_T1, qvar_T1=qvar_T1,

            qET=qET, qEZ_T0=qEZ_T0, qEZ_T1=qEZ_T1,
            weight_views = weight_views
        )

    def initW(self, pmean=0., pvar=1., qmean='random', qvar=1., qE=None, qE2=None, Y=None):
        """Method to initialise the weights

        PARAMETERS
        ----------
        pmean: mean of the prior distribution
        pvar: variance of the prior distribution
        qmean: initial value of the mean of the variational distribution
        qvar: initial value of the variance of the variational distribution
        qE: initial value of the expectation of the variational distribution
        qE2: initial value of the second moment of the variational distribution
        """

        W_list = [None] * self.M

        for m in range(self.M):

            ## Initialise prior distribution (P) ##

            # mean
            pmean_m = np.ones((self.D[m], self.K)) * pmean

            ## Initialise variational distribution (Q) ##

            # variance
            qvar_m = np.ones((self.D[m], self.K)) * qvar

            # mean
            if qmean is not None:
                if isinstance(qmean, str):

                    # Random initialisation
                    if qmean == "random":
                        qmean_m = stats.norm.rvs(loc=0, scale=1., size=(self.D[m], self.K))

                    elif qmean_S1 == "pca":
                        # print("Initialising weights with PCA solution")
                        pca = sklearn.decomposition.PCA(n_components=self.K, whiten=True)
                        pca.fit(Y[m])
                        qmean_S1_tmp = pca.components_.T

                elif isinstance(qmean, np.ndarray):
                    assert qmean.shape == (
                    self.D[m], self.K), "Wrong shape for the expectation of the Q distribution of W"
                    qmean_m = qmean

                elif isinstance(qmean, (int, float)):
                    qmean_m = np.ones((self.D[m], self.K)) * qmean

                else:
                    print("Wrong initialisation for W")
                    exit()

            else:
                qmean_m = None

            # Initialise the node
            W_list[m] = W_Node(
                dim=(self.D[m], self.K),
                pmean=pmean_m, pvar=pvar,
                qmean=qmean_m, qvar=qvar_m,
                qE=qE, qE2=qE2
            )

        self.nodes["W"] = Multiview_Variational_Node(self.M, *W_list)

    def initSW(self, pmean_S0=0., pmean_S1=0., pvar_S0=1., pvar_S1=1., ptheta=1.,
        qmean_S0=0., qmean_S1='random', qvar_S0=1., qvar_S1=1., qtheta=1.,
        qEW_S0=None, qEW_S1=None, qES=None, Y=None):
        """Method to initialise sparse weights with a (reparametrised) spike and slab prior

        PARAMETERS
        ----------
        (...)
        """

        W_list = [None]*self.M

        for m in range(self.M):

            ## Initialise prior distribution (P)

            ## Initialise variational distribution (Q)
            if isinstance(qmean_S1,str):

                if qmean_S1 == "random":
                    qmean_S1_tmp = stats.norm.rvs(loc=0, scale=1., size=(self.D[m],self.K))
                elif qmean_S1 == "pca":
                    # if np.any(np.isnan(Y[m])):
                    #     print("Initialising weights with PCA solution, but data has missing values. Doing quick feature-wise mean imputation (just for the initialisation)... ")
                    #     # TO-DO: BE CAREFUL WITH NON-GAUSSIAN DATA...
                    #     from sklearn.impute import SimpleImputer
                    #     imp = SimpleImputer(missing_values=np.nan, strategy='mean') # using the mean along each column
                    #     imp.fit(Y[m])
                    #     imp.transform(Y[m])

                    pca = sklearn.decomposition.PCA(n_components=self.K, whiten=True)
                    pca.fit(Y[m])
                    qmean_S1_tmp = pca.components_.T
                    # qmean_S1_tmp /= np.nanstd(qmean_S1_tmp, axis=0) # Scale weights to unit variance
                else:
                    print("%s initialisation not implemented for W" % qmean_S1)
                    exit()

                # Scale weights to the variance of the view
                # if Y is not None:
                #     qmean_S1_tmp *= np.nanstd(Y[m]) 

            elif isinstance(qmean_S1,np.ndarray):
                assert qmean_S1.shape == (self.D[m],self.K), "Wrong dimensionality"

            elif isinstance(qmean_S1,(int,float)):
                qmean_S1_tmp = np.ones((self.D[m],self.K)) * qmean_S1

            else:
                print("Wrong initialisation for W")
                exit(1)

            W_list[m] = SW_Node(
                dim=(self.D[m],self.K),

                ptheta=ptheta,
                pmean_S0=pmean_S0, pvar_S0=pvar_S0,
                pmean_S1=pmean_S1, pvar_S1=pvar_S1,

                qtheta=qtheta,
                qmean_S0=qmean_S0, qvar_S0=qvar_S0,
                qmean_S1=qmean_S1_tmp, qvar_S1=qvar_S1,

                qES=qES, qEW_S0=qEW_S0, qEW_S1=qEW_S1
            )

        self.nodes["W"] = Multiview_Variational_Node(self.M, *W_list)

    def initAlphaZ(self, pa=1e-14, pb=1e-14, qa=1., qb=1., qE=None, qlnE=None):
        """Method to initialise the ARD prior on Z per sample group

        PARAMETERS
        ----------
        groups: dictionary
            a dictionariy with mapping between sample names (keys) and samples_groups (values)
        pa: float
            'a' parameter of the prior distribution
        pb :float
            'b' parameter of the prior distribution
        qb: float
            initialisation of the 'b' parameter of the variational distribution
        qE: float
            initial expectation of the variational distribution
        qlnE: float
            initial log expectation of the variational distribution
        """

        self.nodes["AlphaZ"] = AlphaZ_Node(
            dim=(self.G, self.K),
            pa=pa, pb=pb,
            qa=qa, qb=qb,
            groups=self.groups_ix,
            qE=qE, qlnE=qlnE
        )

    def initAlphaW(self, pa=1e-14, pb=1e-14, qa=1., qb=1., qE=None, qlnE=None):
        """Method to initialise the ARD prior on W

        PARAMETERS
        ----------
        pa: float
            'a' parameter of the prior distribution
        pb :float
            'b' parameter of the prior distribution
        qa: float
            initialisation of the 'b' parameter of the variational distribution
        qb: float
            initialisation of the 'b' parameter of the variational distribution
        qE: float
            initial expectation of the variational distribution
        qlnE: float
            initial log expectation of the variational distribution
        """

        alpha_list = [None]*self.M

        for m in range(self.M):
            alpha_list[m] = AlphaW_Node(dim=(self.K,), pa=pa, pb=pb, qa=qa, qb=qb, qE=qE, qlnE=qlnE)
        self.nodes["AlphaW"] = Multiview_Variational_Node(self.M, *alpha_list)

    def initTau(self, pa=1e-3, pb=1e-3, qa=1., qb=1., qE=None):
        """Method to initialise the precision of the noise

        PARAMETERS
        ----------
        pa: float
            'a' parameter of the prior distribution
        pb :float
            'b' parameter of the prior distribution
        qb: float
            initialisation of the 'b' parameter of the variational distribution
        qE: float
            initial expectation of the variational distribution
        """

        tau_list = [None]*self.M

        for m in range(self.M):

            # Poisson noise model for count data
            if self.lik[m] == "poisson":
                tmp = 0.25 + 0.17*np.nanmax(self.data[m],axis=0)
                tmp = np.repeat(tmp[None,:], self.N, axis=0)
                tau_list[m] = Tau_Seeger(dim=(self.N, self.D[m]), value=tmp)

            # Bernoulli noise model for binary data
            elif self.lik[m] == "bernoulli":
                # tau_list[m] = Constant_Node(dim=(self.D[m],), value=np.ones(self.D[m])*0.25)
                # tau_list[m] = Tau_Jaakkola(dim=(self.D[m],), value=0.25)
                tau_list[m] = Tau_Jaakkola(dim=((self.N, self.D[m])), value=1.)

            elif self.lik[m] == "zero_inflated":
                # contains parameters to initialise both normal and jaakola tau
                tau_list[m] = Zero_Inflated_Tau_Jaakkola(dim=((self.G, self.D[m])), value=1., pa=pa, pb=pb, qa=qa, qb=qb, groups=self.groups_ix, qE=qE)

            # Gaussian noise model for continuous data
            elif self.lik[m] == "gaussian":
                tau_list[m] = TauD_Node(dim=(self.G, self.D[m]), pa=pa, pb=pb, qa=qa, qb=qb, groups=self.groups_ix, qE=qE)

        self.nodes["Tau"] = Multiview_Mixed_Node(self.M, *tau_list)

    def initY(self):
        """Method to initialise the observations"""

        Y_list = [None]*self.M
        for m in range(self.M):
            if self.lik[m]=="gaussian":
                Y_list[m] = Y_Node(dim=(self.N,self.D[m]), value=self.data[m], groups=self.groups_ix)
            elif self.lik[m]=="poisson":
                Y_list[m] = Poisson_PseudoY(dim=(self.N,self.D[m]), obs=self.data[m], E=self.data[m], groups=self.groups_ix)
            elif self.lik[m]=="bernoulli":
                Y_list[m] =  Bernoulli_PseudoY_Jaakkola(dim=(self.N,self.D[m]), obs=self.data[m], E=self.data[m], groups=self.groups_ix)
            elif self.lik[m]=="zero_inflated":
                Y_list[m] = Zero_Inflated_PseudoY_Jaakkola(dim=(self.N,self.D[m]), obs=self.data[m], E=self.data[m])
        self.Y = Multiview_Mixed_Node(self.M, *Y_list)
        self.nodes["Y"] = self.Y

    def initThetaZ(self, pa=1., pb=1., qa=1., qb=1., qE=None):
        """Method to initialise the ARD prior on Z per sample group

        PARAMETERS
        ----------
        pa: float
            'a' parameter of the prior distribution
        pb :float
            'b' parameter of the prior distribution
        qb: float
            initialisation of the 'b' parameter of the variational distribution
        qE: float
            initial expectation of the variational distribution
        qlnE: float
            initial log expectation of the variational distribution
        K: number of factors for which we learn theta. If no argument is given, we'll just use the
            total number of factors
        """

        self.nodes["ThetaZ"] = ThetaZ_Node(
            dim=(self.G, self.K),
            pa=pa, pb=pb,
            qa=qa, qb=qb,
            groups=self.groups_ix,
            qE=qE)

    def initThetaW(self, pa=1., pb=1., qa=1., qb=1., qE=None):
        """Method to initialise the sparsity parameter of the spike and slab weights

        PARAMETERS
        ----------
         pa: float
            'a' parameter of the prior distribution
         pb :float
            'b' parameter of the prior distribution
         qb: float
            initialisation of the 'b' parameter of the variational distribution
         qE: float
            initial expectation of the variational distribution
        """
        Theta_list = [None] * self.M
        for m in range(self.M):
            Theta_list[m] = ThetaW_Node(dim=(self.K,), pa=pa, pb=pb, qa=qa, qb=qb, qE=qE)
        self.nodes["ThetaW"] = Multiview_Variational_Node(self.M, *Theta_list)

    def initExpectations(self, *nodes):
        """ Method to initialise all expectations """
        for node in nodes:
            self.nodes[node].updateExpectations()

    def getNodes(self):
        """ Get method to return the nodes """
        return self.nodes
        #return { k:v for (k,v) in self.nodes.items()}



import pandas as pd
import random
import scanpy
import numpy as np
import torch
#from deepforest import CascadeForestClassifier
from sklearn.neighbors import NearestNeighbors
from scipy.linalg import solve
from sklearn.neighbors import KDTree
from scipy.optimize import nnls
import anndata
import os
from collections import defaultdict
import torch.nn as nn
import torch.nn.functional as F
from tqdm import tqdm,trange
from torch.utils.data import TensorDataset, DataLoader



def create_st(generate_sc_data, generate_sc_meta, spot_num, cell_num, gene_num, marker_used):
    sc = generate_sc_data
    sc_ct = generate_sc_meta
    cell_name = sorted(list(set(sc_ct.Cell)))

    last_cell_pool = []
    spots = pd.DataFrame()
    meta = pd.DataFrame(columns=['Cell', 'Celltype', 'Spot'])
    sc_ct.index = sc_ct['Cell']
    for i in range(spot_num):
        cell_pool = random.sample(cell_name, cell_num)
        while set(cell_pool) == set(last_cell_pool):
            cell_pool = random.sample(cell_name, cell_num)
        last_cell_pool = cell_pool
        syn_spot = sc[cell_pool].sum(axis=1)
        if syn_spot.sum() > 25000:
            syn_spot *= 20000 / syn_spot.sum()
        spot_name = f'spot_{i + 1}'
        spots.insert(len(spots.columns), spot_name, syn_spot)

        for cell in cell_pool:
            celltype = sc_ct.loc[cell, 'Cell_type']
            row = {'Cell': cell, 'Celltype': celltype, 'Spot': spot_name}
            #meta = pd.concat([meta,pd.DataFrame(row)],axis=0)
            meta.loc[len(meta)]=pd.Series(row)
            #meta = meta.append(row, ignore_index=True)

    if marker_used:
        adata = scanpy.AnnData(sc.T)

        adata.obs = sc_ct[['Cell_type']]
        scanpy.tl.rank_genes_groups(adata, 'Cell_type', method='wilcoxon')
        marker_df = pd.DataFrame(adata.uns['rank_genes_groups']['names']).head(gene_num)
        marker_array = np.array(marker_df)
        marker_array = np.ravel(marker_array)
        marker_array = np.unique(marker_array)
        marker = list(marker_array)
        sc = sc.loc[marker, :]
        spots = spots.loc[marker, :]
        
    return sc, sc_ct, spots, meta


def create_sample(sc, st, meta, multiple):
    cell_name = meta.Cell.values.tolist()
    spot_name = meta.Spot.values.tolist()

    # get wrong spot name for negative data
    all_spot = list(set(meta.Spot))
    wrong_spot_name = []
    for sn in spot_name:
        last_spot = all_spot.copy()  # --
        last_spot.remove(sn)  # --
        mul_wrong = random.sample(last_spot, multiple)
        wrong_spot_name.extend(mul_wrong)

    cfeat_p_list, cfeat_n_list = [], []

    for c in cell_name:
        cell_feat = sc[c].values.tolist()

        cfeat_p_list.append(cell_feat)
        cfeat_m = [cell_feat * multiple]
        cfeat_n_list.extend(cfeat_m)

    cfeat_p = np.array(cfeat_p_list)  # [n, d]
    cfeat_n = np.array(cfeat_n_list).reshape(-1, cfeat_p.shape[1])  # [n*m, d]

    # positive spot features
    sfeat_p_list = []
    for s in spot_name:
        spot_feat = st[s].values.tolist()
        sfeat_p_list.append(spot_feat)
        # sfeat_p = np.vstack((sfeat_p, spot_feat))
    sfeat_p = np.array(sfeat_p_list)  # [n, d]

    mfeat_p = sfeat_p - cfeat_p
    feat_p = np.hstack((cfeat_p, sfeat_p))
    feat_p = np.hstack((feat_p, mfeat_p))
    print('sucessfully create positive data')

    # negative spot features
    sfeat_n_list = []
    for s in wrong_spot_name:
        spot_feat = st[s].values.tolist()
        sfeat_n = sfeat_n_list.append(spot_feat)
    sfeat_n = np.array(sfeat_n_list)

    mfeat_n = sfeat_n - cfeat_n
    feat_n = np.hstack((cfeat_n, sfeat_n))
    feat_n = np.hstack((feat_n, mfeat_n))
    print('sucessfully create negative data')

    return feat_p, feat_n


def get_data(pos, neg):
    X = np.vstack((pos, neg))
    y = np.concatenate((np.ones(pos.shape[0]), np.zeros(neg.shape[0])))

    return X, y

def create_data_pyomic(single_data:anndata.AnnData,
                       spatial_data:anndata.AnnData,
                       celltype_key:str,spot_key:list=['xcoord','ycoord'],
                       ):
    print("...loading data")
    input_data = {}
    sc_gene=single_data.var.index.values.tolist()
    st_gene=spatial_data.var.index.values.tolist()
    intersection_genes=[]
    for i in sc_gene:
        if i in st_gene:
            intersection_genes.append(i)

    intersect_gene = intersection_genes
    generate_sc_data = single_data[:,intersect_gene].to_df().T
    st_data = spatial_data[:,intersect_gene].to_df().T
    ism=pd.DataFrame(index=single_data.obs.index)
    ism['Cell']=single_data.obs.index
    ism['Cell_type']=single_data.obs[celltype_key].values
    generate_sc_meta = ism
    ssm=pd.DataFrame(index=spatial_data.obs.index)
    ssm['Spot']=spatial_data.obs.index
    ssm['xcoord']=spatial_data.obs[spot_key[0]].values
    ssm['ycoord']=spatial_data.obs[spot_key[1]].values

    input_data["input_sc_meta"] = ism
    input_data["input_sc_data"] = generate_sc_data
    input_data["input_st_data"] = st_data
    input_data["input_st_meta"] = ssm
    input_data["sc_gene"]=sc_gene
    input_data["st_gene"]=st_gene
    input_data["intersect_gene"]=intersect_gene

    ssm=pd.DataFrame(index=spatial_data.obs.index)
    ssm['Spot']=spatial_data.obs.index
    ssm['xcoord']=spatial_data.obs[spot_key[0]].values
    ssm['ycoord']=spatial_data.obs[spot_key[1]].values

    return input_data


    


def create_data(generate_sc_meta, generate_sc_data, st_data, spot_num, cell_num, top_marker_num, marker_used,
                mul_train):
    sc_gene = generate_sc_data.index.values.tolist()
    st_gene = st_data.index.values.tolist()

    intersection_genes=[]
    for i in sc_gene:
        if i in st_gene:
            intersection_genes.append(i)

    intersect_gene = intersection_genes
    generate_sc_data = generate_sc_data.loc[intersect_gene]
    st_data = st_data.loc[intersect_gene]
    sc_train, _, st_train, meta_train = create_st(generate_sc_data, generate_sc_meta,
                                                  spot_num, cell_num,
                                                  top_marker_num, marker_used)
    pos_train, neg_train = create_sample(sc_train, st_train, meta_train, mul_train)
    xtrain, ytrain = get_data(pos_train, neg_train)
    
    return xtrain, ytrain


def creat_pre_data(st, cell_name, spot_name, spot_indx, cfeat, return_np=False):
    spot = spot_name[spot_indx]
    spot_feat = st[spot].values
    tlist = np.isnan(spot_feat).tolist()
    tlist = [i for i, x in enumerate(tlist) if x == True]
    assert len(tlist) == 0

    sfeat = np.tile(spot_feat, (len(cell_name), 1))
    mfeat = sfeat - cfeat
    feat = np.hstack((cfeat, sfeat))
    feat = np.hstack((feat, mfeat))
    if not return_np:
        feat = torch.from_numpy(feat).type(torch.FloatTensor)

    tlist = np.isnan(sfeat).tolist()
    tlist = [i for i, x in enumerate(tlist) if x == True]
    assert len(tlist) == 0

    tlist = np.isnan(cfeat).tolist()
    tlist = [i for i, x in enumerate(tlist) if x == True]
    assert len(tlist) == 0

    return feat

def creat_pre_data_batch(st, cell_name, spot_name, start_indx, end_indx, cfeat, return_np=False):
    feats = []
    for i in range(start_indx, end_indx):
        spot = spot_name[i]
        spot_feat = st[spot].values
        tlist = np.isnan(spot_feat).tolist()
        tlist = [j for j, x in enumerate(tlist) if x == True]
        assert len(tlist) == 0

        sfeat = np.tile(spot_feat, (len(cell_name), 1))
        mfeat = sfeat - cfeat
        feat = np.hstack((cfeat, sfeat))
        feat = np.hstack((feat, mfeat))
        feats.append(feat)

    feats = np.array(feats)
    if not return_np:
        feats = torch.from_numpy(feats).type(torch.FloatTensor)

    tlist = np.isnan(sfeat).tolist()
    tlist = [i for i, x in enumerate(tlist) if x == True]
    assert len(tlist) == 0

    tlist = np.isnan(cfeat).tolist()
    tlist = [i for i, x in enumerate(tlist) if x == True]
    assert len(tlist) == 0

    return feats


def predict_for_one_spot(model, st_test, cell_name, spot_name, spot_indx, cfeat):
    feats = creat_pre_data(st_test, cell_name, spot_name, spot_indx, cfeat, return_np=True)
    outputs = model.predict_proba(feats)[:, 1]
    # outputs = np.where(outputs>0.5, 1, 0)
    predict = outputs.tolist()
    return spot_indx, predict


def predict_for_one_spot_svm(model, st_test, cell_name, spot_name, spot_indx, cfeat):
    feats = creat_pre_data(st_test, cell_name, spot_name, spot_indx, cfeat, return_np=True)
    outputs = model(torch.tensor(feats)).detach().numpy()[:, 1]
    # outputs = np.where(outputs>0.5, 1, 0)
    predict = outputs.tolist()
    return spot_indx, predict

def predict_for_one_spot_net(model, st_test, cell_name, spot_name, spot_indx, cfeat):
    feats = creat_pre_data(st_test, cell_name, spot_name, spot_indx, cfeat, return_np=True)
    outputs = model(torch.tensor(feats)).detach().numpy().reshape(-1)
    # outputs = np.where(outputs>0.5, 1, 0)
    predict = outputs.tolist()
    return spot_indx, predict

def predict_for_one_spot_net_batch(model, st_test, cell_name, spot_name, spot_indx, cfeat):
    feats = creat_pre_data_batch(st_test, cell_name, spot_name, spot_indx, cfeat, return_np=True)
    outputs = model(torch.tensor(feats)).detach().numpy().reshape(-1,len(cell_name))
    return outputs
    # outputs = np.where(outputs>0.5, 1, 0)
    predict = outputs.tolist()
    return spot_indx, predict



# Define the model
class SVM(nn.Module):
    def __init__(self, X_train,num_classes=2):
        super().__init__()
        torch.manual_seed(2)
        self.linear = nn.Linear(X_train.shape[1], num_classes)

    def forward(self, x):
        return self.linear(x)
    
# 定义神经网络模型
class Net(nn.Module):
    def __init__(self,size):
        super(Net, self).__init__()
        torch.manual_seed(2)
        self.linear1 = nn.Linear(size, 256)
        self.relu1 = nn.ReLU()
        self.linear2 = nn.Linear(256, 1)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        x = self.linear1(x)
        x = self.relu1(x)
        x = self.linear2(x)
        x = self.sigmoid(x)
        return x

class DFRunner:
    def __init__(self,
                 generate_sc_data,
                 generate_sc_meta,
                 st_data, st_meta,
                 marker_used,
                 top_marker_num,
                 random_seed=0,
                 n_jobs=1,device=None,):

        self.sc_test_allgene = generate_sc_data  # pandas.DataFrame, generated gene-cell expression data
        self.cell_type = generate_sc_meta  # pandas.DataFrame, cell type
        self.st_data = st_data  # pandas.DataFrame, gene-sport expression data
        self.meta_test = st_meta  # pandas.DataFrame, spot coordinate.
        self.n_jobs=n_jobs

        self.sc_test = self.sc_test_allgene  # pd.DataFrame, test cell-gene expression data.
        self.st_test = self.st_data  # pd.DataFrame, test spot-gene expression data.
        sc_gene = self.sc_test.index.values.tolist()
        st_gene = self.st_test.index.values.tolist()
        
        intersect_gene = list(set(sc_gene).intersection(set(st_gene)))
        self.sc_test = self.sc_test.loc[intersect_gene]
        self.st_test = self.st_test.loc[intersect_gene]
        self.used_device=device
        

        if marker_used:
            print('select top %d marker genes of each cell type...' % top_marker_num)

            sc = scanpy.AnnData(self.sc_test.T)
            sc.obs = self.cell_type[['Cell_type']]
            scanpy.tl.rank_genes_groups(sc, 'Cell_type', method='wilcoxon')
            marker_df = pd.DataFrame(sc.uns['rank_genes_groups']['names']).head(top_marker_num)
            marker_array = np.array(marker_df)
            marker_array = np.ravel(marker_array)
            marker_array = np.unique(marker_array)
            marker = list(marker_array)
            self.sc_test = self.sc_test.loc[marker, :]
            self.st_test = self.st_test.loc[marker, :]

        #self.model = CascadeForestClassifier(random_state=random_seed, n_jobs=n_jobs,
        #                                     verbose=0)
        #self.model = KNN(k=2)
        

        breed = self.cell_type['Cell_type']
        breed_np = breed.values
        breed_set = set(breed_np)
        self.id2label = sorted(list(breed_set))
        self.label2id = {label: idx for idx, label in enumerate(self.id2label)}
        self.cell2label = dict()
        self.label2cell = defaultdict(set)
        for row in self.cell_type.itertuples():
            cell_name = getattr(row, 'Cell')
            cell_type = self.label2id[getattr(row, 'Cell_type')]
            self.cell2label[cell_name] = cell_type
            self.label2cell[cell_type].add(cell_name)

    def run(self, xtrain, ytrain, max_cell_in_diff_spot_ratio, k, save_dir, save_name, load_path=None,
            num_epochs=1000,batch_size=1000,predicted_size=32):
        
        self.model = Net(xtrain.shape[1]).to(self.used_device)
        print('...The model size is %d' % xtrain.shape[1])
        if load_path is None:
            print('...Net training')
            #xtrain = torch.tensor(xtrain, dtype=torch.float32).to(self.used_device)
            #ytrain = torch.tensor(ytrain, dtype=torch.float32).to(self.used_device)
            # 定义损失函数和优化器
            criterion = nn.BCELoss()
            optimizer = torch.optim.SGD(self.model.parameters(), lr=0.001)

            inputs = torch.tensor(xtrain, dtype=torch.float32).to(self.used_device)
            labels = torch.tensor(ytrain.reshape(len(ytrain),1), dtype=torch.float32).to(self.used_device)


            # 将数据分成小批量
            batch_size = batch_size
            dataset = TensorDataset(inputs, labels)
            dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=True)

            # 训练模型
            with trange(num_epochs) as t:
                for epoch in t:
                    running_loss = 0.0
                    num_samples = 0
                    for i, data in enumerate(dataloader):
                        # 从数据集中获取输入和标签
                        inputs, labels = data

                        # 将输入和标签转换为 PyTorch 的变量
                        inputs = inputs.requires_grad_()
                        labels = labels.float()

                        # 计算模型输出和损失
                        outputs = self.model(inputs)
                        loss = criterion(outputs, labels)

                        # 反向传播和优化
                        optimizer.zero_grad()
                        loss.backward()
                        optimizer.step()

                        # 记录损失值
                        running_loss += loss.item()*inputs.size(0)
                        num_samples += inputs.size(0)

                    # 输出平均损失值
                    average_loss = running_loss / num_samples
                    t.set_description('Epoch [{}/{}], Loss: {:.4f}'.format(epoch+1, num_epochs, average_loss))

            '''  
            # 训练模型
            num_epochs = 10000
            for epoch in range(num_epochs):
                # 计算模型输出和损失
                output_tensor = self.model(inputs)
                loss = criterion(output_tensor, labels)

                # 反向传播和优化
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                
                if epoch % 100 == 0:
                    print('Epoch [{}/{}], Loss: {:.4f}'.format(epoch+1, num_epochs, loss.item()))
            '''
            """
            self.model = SVM(xtrain,2).to(self.used_device)

            # Define the loss function
            criterion = nn.MultiMarginLoss()

            # Define the optimizer
            optimizer = torch.optim.SGD(self.model.parameters(), lr=0.001)

            # Train the model
            for epoch in range(10000):
                optimizer.zero_grad()
                y_pred = self.model(xtrain)
                loss = criterion(y_pred, ytrain)
                loss.backward()
                optimizer.step()
                if epoch % 100 == 0:
                    print("Epoch: %d, Loss: %.4f" % (epoch, loss.item()))
"""
            #self.model.fit(xtrain, ytrain)  # train model
            print('...Net training done!')
            
            path_save = os.path.join(save_dir, f"{save_name}.pth")
            if not os.path.exists(save_dir):
                os.makedirs(save_dir)
            torch.save(self.model.state_dict(), path_save)
            print(f"...save trained net in {path_save}.")

        else:
            self.model.load_state_dict(torch.load(load_path,map_location=self.used_device))
            #self._load_model(load_path)

        #return self.cre_csv(max_cell_in_diff_spot_ratio, k)
        df_meta, df_spot = self.cre_csv(max_cell_in_diff_spot_ratio, k,predicted_size)  # save predicted results.
        return df_meta, df_spot

    def load(self, load_path,modelsize,max_cell_in_diff_spot_ratio, k,predicted_size):
        
        self.model = Net(modelsize).to(self.used_device)
        self.model.load_state_dict(torch.load(load_path,map_location=self.used_device))
        #self._load_model(load_path

        df_meta, df_spot = self.cre_csv(max_cell_in_diff_spot_ratio, k,predicted_size)  # save predicted results.
        return df_meta, df_spot


    def cre_csv(self, max_cell_in_diff_spot_ratio, k,batch_size):
        # data
        cell_name = self.sc_test.columns.values.tolist()
        spot_name = self.st_test.columns.tolist()  # list of spot name ['spot_1', 'spot_2', ...]
        cfeat = self.sc_test.values.T
        cell_num = len(cell_name)
        spot_num = len(spot_name)
        batch_size=batch_size
        if max_cell_in_diff_spot_ratio is None:
            max_cell_in_diff_spot = None
        else:
            max_cell_in_diff_spot = int(max_cell_in_diff_spot_ratio * k * spot_num / cell_num)

        def joint_predict(ratio):            
            score_triple_list = list()
            spot2cell = defaultdict(set)
            cell2spot = defaultdict(set)
            spot2ratio = dict()
            
            print('Calculating scores...')
            #batch_size = batch_size  # adjust this value based on the available memory and the size of the data
            batch_list = [(i, min(i+batch_size, spot_num)) for i in range(0, spot_num, batch_size)]

            with trange(len(batch_list)) as t:
                for batch_indx in t:
                    start_indx, end_indx = batch_list[batch_indx]
                    feat=creat_pre_data_batch(self.st_test, cell_name, spot_name, start_indx,end_indx, cfeat, return_np=True)
                    predict=self.model(torch.tensor(feat, dtype=torch.float32).to(self.used_device)).cpu().detach().numpy().reshape(-1,len(cell_name))
                    for spot_indx,spot_p_num in zip(range(start_indx, end_indx),range(len(spot_name[start_indx:end_indx]))):
                        spot = spot_name[spot_indx]  # spotname
                        for c, p in zip(cell_name,predict[spot_p_num]):
                            score_triple_list.append((c, spot, p))
                        spot2ratio[spot] = np.round(ratio[spot_indx] * k)
                        t.set_description('Now calculating scores for spot %d/%d' % (spot_indx + 1, len(spot_name)))

            print('Calculating scores done.')

            # spot2ratio: map spot to cell type ratio in it.
            score_triple_list = sorted(score_triple_list, key=lambda x: x[2], reverse=True)
            # sort by score
            for c, spt, score in score_triple_list:
                # cell name, spot name, score
                if max_cell_in_diff_spot is not None and len(cell2spot[c]) == max_cell_in_diff_spot:
                    # The number of this cell in different spots reaches a maximum
                    continue
                if len(spot2cell[spt]) == k:
                    # The maximum number of cells in this spot
                    continue
                cell_class = self.cell2label.get(c)
                if cell_class is None:
                    continue
                if spot2ratio[spt][cell_class] > 0:
                    # Put this cell in this spot
                    spot2ratio[spt][cell_class] -= 1
                    spot2cell[spt].add(c)
                    cell2spot[c].add(spt)
                else:
                    continue

            cell_list, spot_list, spot_len = list(), list(), list()
            df_spots = pd.DataFrame()

            order_list = spot_name
            for spot in order_list:
                if spot2cell.get(spot):
                    cells = spot2cell.get(spot)
                    cell_num = len(cells)
                    cell_list.extend(sorted(list(cells)))
                    spot_list.extend([spot] * cell_num)
                    spot_len.append(cell_num)
                    cell_pool = list(cells)
                    cell_pool.sort()

                    predict_spot = self.sc_test_allgene[cell_pool]
                    df_spots = pd.concat([df_spots, predict_spot], axis=1)
            return cell_list, spot_list, spot_len, df_spots

        ratio = self.__calc_ratio()  # [spot_num, class_num]

        cell_list, spot_list, spot_len, df_spots = joint_predict(ratio)
        #return joint_predict(ratio)
        meta = {'Cell': cell_list, 'Spot': spot_list}
        df = pd.DataFrame(meta)
        self.cell_type = self.cell_type.reset_index(drop=True)
        df_meta = pd.merge(df, self.cell_type, how='left')
        df_meta = df_meta[['Cell', 'Cell_type', 'Spot']]
        df_meta = pd.merge(df_meta, self.meta_test, how='inner')

        df_meta = df_meta.rename(columns={'xcoord': 'Spot_xcoord', 'ycoord': 'Spot_ycoord'})

        coord = self.meta_test[['xcoord', 'ycoord']].to_numpy()
        nbrs = NearestNeighbors(n_neighbors=2, algorithm='ball_tree').fit(coord)
        distances, indices = nbrs.kneighbors(coord)
        radius = distances[:, -1] / 2
        radius = radius.tolist()
        all_coord = df_meta[['Spot_xcoord', 'Spot_ycoord']].to_numpy()
        all_radius = list()

        for i in range(len(spot_len)):
            a = [radius[i]] * spot_len[i]
            all_radius.extend(a)

        length = np.random.uniform(0, all_radius)
        angle = np.pi * np.random.uniform(0, 2, all_coord.shape[0])

        x = all_coord[:, 0] + length * np.cos(angle)
        y = all_coord[:, 1] + length * np.sin(angle)
        cell_coord = {'Cell_xcoord': np.around(x, 2).tolist(), 'Cell_ycoord': np.around(y, 2).tolist()}
        df_cc = pd.DataFrame(cell_coord)
        df_meta = pd.concat([df_meta, df_cc], axis=1)

        cell_rename = [f'C_{i}' for i in range(1, df_spots.shape[1] + 1)]
        df_spots.columns = cell_rename
        df_meta = df_meta.drop(['Cell'], axis=1)
        df_meta.insert(0, "Cell", cell_rename)

        return df_meta, df_spots

    def __calc_ratio(self):

        label_devide_data = dict()
        for label, cells in self.label2cell.items():
            label_devide_data[label] = self.sc_test[list(cells)]

        single_cell_splitby_breed_np = {}
        for key in label_devide_data.keys():
            single_cell_splitby_breed_np[key] = label_devide_data[key].values  # [gene_num, cell_num]
            single_cell_splitby_breed_np[key] = single_cell_splitby_breed_np[key].mean(axis=1)

        max_decade = len(single_cell_splitby_breed_np.keys())
        single_cell_matrix = []
        
        for i in range(max_decade):
            single_cell_matrix.append(single_cell_splitby_breed_np[i].tolist())

        single_cell_matrix = np.array(single_cell_matrix)
        single_cell_matrix = np.transpose(single_cell_matrix)  # (gene_num, label_num)

        num_spot = self.st_test.values.shape[1]

        spot_ratio_values = np.zeros((num_spot, max_decade))  # (spot_num, label_num)
        spot_values = self.st_test.values  # (gene_num, spot_num)

        for i in range(num_spot):
            ratio_list = [0 for x in range(max_decade)]
            spot_rep = spot_values[:, i].reshape(-1, 1)
            spot_rep = spot_rep.reshape(spot_rep.shape[0], )

            ratio = nnls(single_cell_matrix, spot_rep)[0]

            ratio_list = [r for r in ratio]
            ratio_list = (ratio_list / np.sum([ratio_list], axis=1)[0]).tolist()

            for j in range(max_decade):
                spot_ratio_values[i, j] = ratio_list[j]
                
        return spot_ratio_values

    def _load_model(self, save_path):
        self.model.load(save_path)
        print(f"loading model from {save_path}")

    def _save_model(self, save_path):
        self.model.save(save_path)
        print(f"saving model to {save_path}")
        return save_path


def aprior(gamma_hat, axis=None):
    m = np.mean(gamma_hat, axis=axis)
    s2 = np.var(gamma_hat, ddof=1, axis=axis)
    
    return (2 * s2 + np.power(m, 2)) / s2


def bprior(gamma_hat, axis=None):
    m = np.mean(gamma_hat, axis=axis)
    s2 = np.var(gamma_hat, ddof=1, axis=axis)
    
    return (m * s2 + np.power(m, 3)) / s2


def postmean(g_hat, g_bar, n, d_star, t2):
    return (t2 * n * g_hat + d_star * g_bar) / (t2 * n + d_star)


def postvar(sum2, n, a, b):
    return (0.5 * sum2 + b) / (n / 2 + a - 1)


def it_sol(sdat, g_hat, d_hat, g_bar, t2, a, b, conv=0.0001):
    n = np.sum(~np.isnan(sdat), axis=1)
    g_old = g_hat
    d_old = d_hat
    change = 1
    count = 0
    while change > conv:
        g_new = postmean(g_hat, g_bar, n, d_old, t2)
        sum2 = np.nansum(np.power(sdat - np.dot(np.expand_dims(g_new, axis=1), np.ones((1, sdat.shape[1]))), 2), axis=1)
        d_new = postvar(sum2, n, a, b)
        change = np.max((np.max(np.abs(g_new - g_old) / g_old), np.max(np.abs(d_new - d_old) / d_old)))
        g_old = g_new
        d_old = d_new
        count += 1
        
    return np.concatenate((np.expand_dims(g_new, axis=1), np.expand_dims(d_new, axis=1)), axis=1)

def joint_analysis(dat, batch, mod=None, par_prior=True, proir_plots=False, mean_only=False, ref_batch=None):
    rownames = dat.index
    colnames = dat.columns
    dat = np.array(dat)
    batch_levels = batch.drop_duplicates()

    batches = []
    ref_index = 0
    zero_rows_list = []
    for i, batch_level in enumerate(batch_levels):
        idx = batch.isin([batch_level])
        if batch_level == ref_batch:
            ref_index = i
        batches.append(idx.reset_index().loc[lambda d: d.Batch == True].index)
        batch_dat = dat[:, idx]
        for row in range(np.size(batch_dat, 0)):
            if np.var(batch_dat[row, :], ddof=1) == 0:
                zero_rows_list.append(row)
    zero_rows = list(set(zero_rows_list))
    keep_rows = list(set(range(dat.shape[0])).difference(set(zero_rows_list)))
    dat_origin = dat
    dat = dat[keep_rows, :]
    batchmod = pd.get_dummies(batch, drop_first=False, prefix='batch')
    batchmod['batch_' + ref_batch] = 1
    ref = batchmod.columns.get_loc('batch_' + ref_batch)
    design = np.array(batchmod)
    n_batch = batch_levels.shape[0]
    n_batches = [len(x) for x in batches]

    B_hat = solve(np.dot(design.T, design), np.dot(design.T, dat.T))
    grand_mean = np.expand_dims(B_hat[ref, :], axis=1)
    ref_dat = dat[:, batches[ref_index]]
    var_pool = np.expand_dims(np.dot(np.square(ref_dat - np.dot(design[batches[ref_index], :], \
                                                                B_hat).T),
                                     np.ones(n_batches[ref_index]).T * 1 / n_batches[ref_index]), axis=1)
    stand_mean = np.dot(grand_mean, np.ones((1, batch.shape[0])))
    s_data = (dat - stand_mean) / np.dot(np.sqrt(var_pool), np.ones((1, batch.shape[0])))
    batch_design = design
    gamma_hat = solve(np.dot(batch_design.T, batch_design), np.dot(batch_design.T, s_data.T))
    delta_hat = np.empty([0, s_data.shape[0]])
    for i in batches:
        row_vars = np.expand_dims(np.nanvar(s_data[:, i], axis=1, ddof=1), axis=0)
        delta_hat = np.concatenate((delta_hat, row_vars), axis=0)
    gamma_bar = np.mean(gamma_hat, axis=1)
    t2 = np.var(gamma_hat, axis=1, ddof=1)
    a_prior = aprior(delta_hat, axis=1)
    b_prior = bprior(delta_hat, axis=1)
    results = []
    gamma_star = np.empty((n_batch, s_data.shape[0]))
    delta_star = np.empty((n_batch, s_data.shape[0]))
    for j, batch_level in enumerate(batch_levels):
        i = batchmod.columns.get_loc('batch_' + batch_level)
        results.append(it_sol(s_data[:, batches[j]], gamma_hat[i, :], delta_hat[j, :], gamma_bar[i], t2[i], a_prior[j],
                              b_prior[j]).T)
    for j, batch_level in enumerate(batch_levels):
        gamma_star[j, :] = results[j][0]
        delta_star[j, :] = results[j][1]
    gamma_star[ref_index, :] = 0
    delta_star[ref_index, :] = 1
    bayesdata = s_data
    for i, batch_index in enumerate(batches):
        bayesdata[:, batch_index] = (bayesdata[:, batch_index] - np.dot(batch_design[batch_index, :], gamma_star).T) / \
                                    np.dot(np.sqrt(np.expand_dims(delta_star[i], axis=1)), np.ones((1, n_batches[i])))

    bayesdata = (bayesdata * np.dot(np.sqrt(var_pool), np.ones((1, dat.shape[1])))) + stand_mean
    bayesdata[:, batches[ref_index]] = dat[:, batches[ref_index]]
    if len(zero_rows) > 0:
        dat_origin[keep_rows, :] = bayesdata
        bayesdata = pd.DataFrame(dat_origin, index=rownames, columns=colnames)
    bayesdata[bayesdata < 0] = 0
    
    return bayesdata


def knn(data, query, k):
    tree = KDTree(data)
    dist, ind = tree.query(query, k)
    
    return dist, ind



import numpy as np
import pandas as pd
import anndata
import scanpy as sc


from sklearn.model_selection import train_test_split


def dynamical_analysis(sc_object,sc_object_aggr, n_states = None, n_states_seq = None,
                        weight_connectivities=0.2, n_components=20, thresh_ms_gene = 0, 
                        use_spatial = False, spa_weight = 0.5, spa_conn_key = 'spatial',
                        n_jobs=-1):
    """
    Perform STT dynamical analysis on a single-cell transcriptomics dataset.

    Parameters:
    -----------
    sc_object : AnnData object
        Annotated data matrix with rows for cells and columns for genes.
    sc_object_aggr : AnnData object
        Annotated data matrix with rows for cell aggregates and columns for genes.
    n_states : int, optional
        Number of macrostates to compute. If None, n_states_seq must be provided.
    n_states_seq : list of int, optional
        Sequence of number of macrostates to compute. If None, n_states must be provided.
    weight_connectivities : float, optional
        Weight of gene expression similarity connectivities in computing the transition matrix.
    n_components : int, optional
        Number of components to compute in the Schur decomposition.
    thresh_ms_gene : float, optional
        Threshold for selecting genes based on their mean squared expression.
    use_spatial : bool, optional
        Whether to use spatial information in computing the transition matrix.
    spa_weight : float, optional
        Weight of spatial similarity in computing the transition matrix.
    spa_conn_key : str, optional
        Key for accessing the spatial connectivities in the AnnData object.

    Returns:
    --------
    None, but the results are stored in the AnnData object including the transition matrix, the stationary distribution, the macrostates memberships, the selected genes, and the kernel.
    """
    import cellrank as cr

    from cellrank.estimators import GPCCA
    gene_select = sc_object.var['r2_test'][sc_object.var['r2_test']>thresh_ms_gene].index.tolist()
    gene_subset = [gene+'_u' for gene in gene_select]+gene_select
    r2_keep_test = sc_object.var['r2_test'][sc_object.var['r2_test']>thresh_ms_gene]
    r2_keep_train = sc_object.var['r2_train'][sc_object.var['r2_test']>thresh_ms_gene]   
    


    #kernel_tensor = cr.tl.transition_matrix(sc_object_aggr, weight_connectivities=0,  n_jobs=-1,scheme = 'dot_product',gene_subset = gene_subset)
    kernel_tensor = cr.kernels.VelocityKernel(sc_object_aggr, gene_subset = gene_subset)
    kernel_tensor.compute_transition_matrix(n_jobs=n_jobs,show_progress_bar = False,similarity = 'dot_product')

    kernel = (1-weight_connectivities)*kernel_tensor+weight_connectivities*sc_object.uns['kernel_connectivities']
    if use_spatial:
        kernel = (1-spa_weight)*kernel + spa_weight*sc_object.uns['kernel_spatial']
    
    g_fwd = GPCCA(kernel)
    g_fwd.compute_schur(n_components=n_components)
    
    if n_states == None:
        g_fwd.compute_macrostates(n_states=n_states_seq,use_min_chi = True)
    else:
        g_fwd.compute_macrostates(n_states=n_states)    
    
    P_cg = g_fwd.coarse_T.to_numpy()
    P_cg[P_cg<0] = 1e-16
    row_sums = P_cg.sum(axis=1)
    P_hat = P_cg / row_sums[:, np.newaxis]
    
    sc_object.uns['da_out'] = {}
    sc_object.uns['da_out']['P_hat'] = P_hat
    sc_object.uns['da_out']['mu_hat'] = g_fwd.coarse_stationary_distribution.to_numpy()
    sc_object.uns['da_out']['membership'] = g_fwd.macrostates_memberships.X
    sc_object.uns['da_out']['gene_select'] = gene_select
    sc_object.uns['kernel'] = kernel
    sc_object.obsm['rho'] = g_fwd.macrostates_memberships.X

    
    sc_object_aggr.obs['attractor'] = sc_object.obs['attractor'].values
    sc_object_aggr.uns['gene_subset'] = gene_subset
    sc_object.uns['r2_keep_train'] = r2_keep_train
    sc_object.uns['r2_keep_test'] = r2_keep_test



def construct_tenstor(adata, rho, portion = 0.8,l=0):
    """ 
    Construct the tensor for the dynamical analysis.
    
    Parameters:
    -----------  
    adata: AnnData object
    rho: tensor N_c*N_g*K
    portion: float, optional (default: 0.8)
        The portion of the data to be used for training.    
    l: float, optional (default: 0)
        The regularization parameter.
    
    Returns:
    --------
    None, but the tensor is stored in adata.obsm['tensor_v'] and the parameters are stored in adata.uns['par'] 
    """
    # tensor N_c*N_g*2*K
    K = rho.shape[1]
    par = np.zeros((adata.shape[1],K+1))
    tensor_v = np.zeros((adata.shape[0],adata.shape[1],2,K))
    r2_train = np.zeros(adata.shape[1])
    r2_test = np.zeros(adata.shape[1])
 
    U = adata.layers['Mu']
    S = adata.layers['Ms']
    if 'toarray' in dir(U):
        U = U.toarray()
        S = S.toarray()

    label_all = adata.obs['attractor']
    categories = np.unique(label_all)
    U_train, U_test, S_train, S_test, rho_train, rho_test,label,_ = train_test_split(U, S, rho, label_all, test_size=1-portion, random_state=42)

    for i in range(adata.shape[1]):
        u_train = U_train[:,i]
        s_train = S_train[:,i]
        u_test = U_test[:,i]
        s_test = S_test[:,i]
        
        if 'toarray' in dir(u_train):
            u_train = u_train.toarray().flatten()
            s_train = s_train.toarray().flatten()
        
        indices_per_category = []

        for category in categories:
            cat_indices = np.where(label == category)[0]
            u_train_cat = u_train[cat_indices]
            s_train_cat = s_train[cat_indices]
            if len(u_train_cat[u_train_cat>0])>0:
                u_q = u_train_cat[u_train_cat>0]
            else:
                u_q = u_train_cat
            if len(s_train_cat[s_train_cat>0])>0:
                s_q = s_train_cat[s_train_cat>0]
            else:
                s_q = s_train_cat

            u_10, u_90 = np.quantile(u_q, [0.1, 0.9])
            s_10, s_90 = np.quantile(s_q, [0.1, 0.9])
            selected_indices = cat_indices[np.where((u_train_cat >= u_10) & (u_train_cat <= u_90) & (s_train_cat >= s_10) & (s_train_cat <= s_90))[0]]
            indices_per_category.append(selected_indices)

        indices = np.unique(np.concatenate(indices_per_category))
        if len(indices>=10):
            # Extract the values from u_train and s_train using the indices
            u_train_selected = u_train[indices]
            s_train_selected = s_train[indices]
            rho_train_selected = rho_train[indices,:]

            m_c = np.zeros(K)
            U_c_var = np.zeros(u_train_selected.shape[0])
            
            for c in range(K):
                if np.max(rho_train_selected[:,c])>0:
                    m_c[c] =np.inner(u_train_selected,rho_train_selected[:,c])/(np.sum(rho_train_selected[:,c])+l) #np.average(u_train_selected,weights = rho_train_selected[:,c])
                else:
                    m_c[c] = 0

            
            for k in range(u_train_selected.shape[0]):
                U_c_var[k] = np.inner((u_train_selected[k]-m_c)**2,rho_train_selected[k,:])
            
            par[i,K]= np.inner(u_train_selected,s_train_selected)/(np.sum(u_train_selected**2+U_c_var)+l)  #beta
            par[i,:K] = m_c*par[i,K] #alpha
            
            U_beta_train = par[i,K]*u_train
            U_beta_test = par[i,K]*u_test
            var_reg_train = np.sum((par[i,K]*u_train-s_train)**2)+np.sum(((U_beta_train[:,np.newaxis]-par[i,:K])**2)*rho_train)
            var_reg_test = np.sum((par[i,K]*u_test-s_test)**2)+np.sum(((U_beta_test[:,np.newaxis]-par[i,:K])**2)*rho_test)
            var_all_train = U_train.shape[0]*(np.var(s_train)+np.var(U_beta_train))
            var_all_test = U_test.shape[0]*(np.var(s_test)+np.var(U_beta_test))
            
            r2_train[i] = 1- var_reg_train/var_all_train
            r2_test[i] = 1- var_reg_test/var_all_test
        else:
            r2_train[i] = -100
            r2_test[i] = -100

        

    for i in range(K):        
        tensor_v [:,:,0,i] = (par[:,i]- U * par[:,K])
        tensor_v [:,:,1,i] = (U * par[:,K] - S)
        
        
    adata.uns['par'] = par
    adata.obsm['tensor_v'] = tensor_v
    adata.var['r2_train'] = r2_train
    adata.var['r2_test'] = r2_test
    
def aver_velo(tensor_v,membership):
    (N_c,N_g,_,_) = np.shape(tensor_v)
    velo = np.zeros((N_c,N_g,2))
    
    for i in range(N_c):
        for j in range(N_g):
            for k in range(2):
                velo[i,j,k] = np.dot(tensor_v[i,j,k,:] ,membership[i,:])
    
    return velo

def dynamical_iteration(adata, n_states=None, n_states_seq=None, 
                        n_iter=10, return_aggr_obj=True, weight_connectivities=0.2, 
                        n_components=20, n_neighbors=100, thresh_ms_gene=0, 
                        thresh_entropy=0.1, use_spatial=False, spa_weight=0.5, 
                        spa_conn_key='spatial', monitor_mode=False, l2=0.1,n_jobs=-1):
    """
    Perform dynamical iteration on the given AnnData object. The function updates the attractor states, the tensor, the averaged velocity, and the entropy at each iteration.

    Parameters:
    -----------
    adata: AnnData object
        Annotated data matrix.
    n_states: int, optional (default: None)
        Number of attractor states.
    n_states_seq: list, optional (default: None)
        List of number of attractor states for each iteration.
    n_iter: int, optional (default: 10)
        Number of iterations.
    return_aggr_obj: bool, optional (default: False)
        Whether to return the aggregated object.
    weight_connectivities: float, optional (default: 0.2)
        Weight of connectivities.
    n_components: int, optional (default: 20)
        Number of components.
    n_neighbors: int, optional (default: 100)
        Number of neighbors.
    thresh_ms_gene: int, optional (default: 0)
        Threshold for mean spliced gene expression.
    thresh_entropy: float, optional (default: 0.1)
        Threshold for entropy.
    use_spatial: bool, optional (default: False)
        Whether to use spatial information.
    spa_weight: float, optional (default: 0.5)
        Weight of spatial information.
    spa_conn_key: str, optional (default: 'spatial')
        Key for spatial connectivities.
    stop_cr: str, optional (default: 'abs')
        Stopping criterion for iteration.
    monitor_mode: bool, optional (default: False)
        Whether to use monitor mode.
    l2: float, optional (default: 0.1)
        Regularization parameter in tensor estimation.

    Returns:
    --------
    Default none, but updates the adata.uns with the following keys: 
    da_out: dict
        Dictionary of the results of dynamical analysis.
    gene_subset: list
        List of selected multi-stable genes.
    entropy: np.ndarray
        Array of entropy values.
    speed: np.ndarray
        Array of speed values.
    attractor: np.ndarray
        Array of attractor states.
    tensor_v_aver: np.ndarray
        Array of averaged tensor by attractor membership of cells.
        
    If return_aggr_obj is True, the aggregated object with both spliced and unspliced counts of multi-stable genes is returned.

    """
    from cellrank.kernels import ConnectivityKernel
    adata.uns['da_out']={}
    rho = pd.get_dummies(adata.obs['attractor']).to_numpy()
    construct_tenstor(adata,rho = rho)
    adata.obsm['tensor_v_aver'] = aver_velo(adata.obsm['tensor_v'],rho) 
    adata.obsm['rho'] =rho 
    kernel_similarity = ConnectivityKernel(adata)
    kernel_similarity.compute_transition_matrix(density_normalize = False)
    adata.uns['kernel_connectivities'] = kernel_similarity
    if use_spatial:
        spa_kernel = ConnectivityKernel(adata,conn_key=spa_conn_key+'_connectivities')
        spa_kernel.compute_transition_matrix()
        adata.uns['kernel_spatial'] = spa_kernel
    
    U = adata.layers['unspliced']
    S = adata.layers['spliced']

    if 'toarray' in dir(U):
        U = U.toarray()
        S = S.toarray()
    
    X_all = np.concatenate((U,S),axis = 1)
    
    sc_object_aggr = anndata.AnnData(X=X_all)
    sc_object_aggr.var_names = [gene+'_u' for gene in adata.var_names.to_list()]+adata.var_names.to_list()
    
    velo = adata.obsm['tensor_v_aver']
    sc_object_aggr.layers['velocity']= np.concatenate((velo[:,:,0],velo[:,:,1]),axis = 1)
    sc_object_aggr.layers['spliced'] = X_all
    sc_object_aggr.layers['Ms']=sc_object_aggr.layers['spliced']
    
    sc_object_aggr.obs['entropy'] = -np.sum(rho*np.log(rho+1e-8),axis = 1)
    entropy = 0
    sc_object_aggr.var['highly_variable'] = True
    
    for i in range(n_iter):
        entropy_orig = entropy
        
        sc.tl.pca(sc_object_aggr,use_highly_variable = True)
        sc.pp.neighbors(sc_object_aggr,n_neighbors = n_neighbors)#update the neighbors using multistable genes
        dynamical_analysis(adata, sc_object_aggr, n_states = n_states,
                           n_states_seq=n_states_seq, n_jobs=n_jobs,
                           weight_connectivities=weight_connectivities, 
                           n_components = n_components,thresh_ms_gene = thresh_ms_gene, 
                           use_spatial = use_spatial, spa_weight = spa_weight, spa_conn_key = spa_conn_key)
        
        rho = adata.obsm['rho']
        adata.obs['attractor'] =  np.argmax(rho,axis = 1)
        adata.obs['attractor'] = adata.obs['attractor'].astype('category')
        
        construct_tenstor(adata,rho = rho,l=l2)
        adata.obsm['tensor_v_aver'] = aver_velo(adata.obsm['tensor_v'],rho)
        sc_object_aggr.layers['velocity']= np.concatenate((adata.obsm['tensor_v_aver'][:,:,0],adata.obsm['tensor_v_aver'][:,:,1]),axis = 1)
        sc_object_aggr.var['highly_variable'] = [genes in sc_object_aggr.uns['gene_subset'] for genes in sc_object_aggr.var_names]
        
        
        entropy = -np.sum(rho*np.log(rho+1e-8),axis = 1)
        ent_diff_rel = np.abs(entropy-entropy_orig)/np.abs(entropy_orig)
        err_diff_abs = np.abs(entropy-entropy_orig)
        ent_diff_rel_series = pd.Series(ent_diff_rel)
        err_diff_abs_series = pd.Series(err_diff_abs)

        quantiles = [0, 0.25, 0.5, 0.75, 1]
        print("\nQuantiles for entropy relative difference with last iteration:")
        print(ent_diff_rel_series.quantile(quantiles))

        print("\nQuantiles for entropy absolute difference with last iteration:")
        print(err_diff_abs_series.quantile(quantiles))

        err_ent = np.quantile(ent_diff_rel,0.75)
                
        if monitor_mode:
            user_input = input("Do you want to continue? (y/n): ").strip().lower()
            if user_input == 'n':
                    print("Exiting the loop.")
                    break
            elif user_input != 'y':
                    print("Invalid input, please enter 'y' to continue or 'n' to exit.")
            
            print(f"Current values:\n weight_connectivities: {weight_connectivities}, thresh_ms_gene: {thresh_ms_gene}, spa_weight: {spa_weight}")

            # Ask for updates
            update_choice = input("Do you want to update parameters? (y/n): ").lower()
            if update_choice == 'y':
                # Ask for each parameter and update if the user wishes to change it
                new_weight_connectivities = input(f"Enter new value for weight_connectivities (current: {weight_connectivities}) or press Enter to keep current: ")
                weight_connectivities = float(new_weight_connectivities) if new_weight_connectivities else weight_connectivities

                new_thresh_ms_gene = input(f"Enter new value for thresh_ms_gene (current: {thresh_ms_gene}) or press Enter to keep current: ")
                thresh_ms_gene = float(new_thresh_ms_gene) if new_thresh_ms_gene else thresh_ms_gene

                new_spa_weight = input(f"Enter new value for spa_weight (current: {spa_weight}) or press Enter to keep current: ")
                spa_weight = float(new_spa_weight) if new_spa_weight else spa_weight

        else:
            err_ent = np.quantile(ent_diff_rel,0.75)
            if err_ent < thresh_entropy:
                print("Entropy difference is below the threshold, exiting the loop.")
                break


    adata.obs['entropy'] = entropy
    adata.uns['gene_subset'] = sc_object_aggr.uns['gene_subset']
    sc_object_aggr.layers['Ms'] = np.concatenate((adata.layers['Mu'],adata.layers['Ms']),axis = 1)
    sc_object_aggr.layers['vj'] = np.concatenate((adata.obsm['tensor_v_aver'][:,:,0],adata.obsm['tensor_v_aver'][:,:,1]),axis = 1)


    sc_object_aggr = sc_object_aggr[:,sc_object_aggr.uns['gene_subset']]
    adata.obs['speed'] = np.linalg.norm(sc_object_aggr.layers['velocity'],axis=1)
    adata.layers['velo'] = adata.obsm['tensor_v_aver'][:,:,1]
    
    if return_aggr_obj:
        sc_object_aggr.obs['entropy'] = adata.obs['entropy'].values
        sc_object_aggr.obs['speed'] = adata.obs['speed'].values
        sc_object_aggr.obs['attractor'] = adata.obs['attractor'].values
        return sc_object_aggr
    


import numpy as np
import scanpy as sc
import pandas as pd

import io
import anndata

def ligand_receptor_database(
    database = "CellChat",
    species = "mouse",
    heteromeric_delimiter = "_",
    signaling_type = "Secreted Signaling" # or "Cell-Cell contact" or "ECM-Receptor" or None
):
    """
    Extract ligand-receptor pairs from LR database.

    Parameters
    ----------
    database
        The name of the ligand-receptor database. Use 'CellChat' for CellChatDB [Jin2021]_ of 'CellPhoneDB_v4.0' for CellPhoneDB_v4.0 [Efremova2020]_.
    species
        The species of the ligand-receptor pairs. Choose between 'mouse' and 'human'.
    heteromeric_delimiter
        The character to separate the heteromeric units of heteromeric ligands and receptors. 
        For example, if the heteromeric receptor (TGFbR1, TGFbR2) will be represented as 'TGFbR1_TGFbR2' if this parameter is set to '_'.
    signaling_type
        The type of signaling. Choose from 'Secreted Signaling', 'Cell-Cell Contact', and 'ECM-Receptor' for CellChatDB or 'Secreted Signaling' and 'Cell-Cell Contact' for CellPhoneDB_v4.0. 
        If None, all pairs in the database are returned.

    Returns
    -------
    df_ligrec : pandas.DataFrame
        A pandas DataFrame of the LR pairs with the three columns representing the ligand, receptor, and the signaling pathway name, respectively.

    References
    ----------

    .. [Jin2021] Jin, S., Guerrero-Juarez, C. F., Zhang, L., Chang, I., Ramos, R., Kuan, C. H., ... & Nie, Q. (2021). 
        Inference and analysis of cell-cell communication using CellChat. Nature communications, 12(1), 1-20.
    .. [Efremova2020] Efremova, M., Vento-Tormo, M., Teichmann, S. A., & Vento-Tormo, R. (2020). 
        CellPhoneDB: inferring cell–cell communication from combined expression of multi-subunit ligand–receptor complexes. Nature protocols, 15(4), 1484-1506.

    """
    import pkgutil
    if database == "CellChat":
        data = pkgutil.get_data(__name__, "_data/LRdatabase/CellChat/CellChatDB.ligrec."+species+".csv")
        df_ligrec = pd.read_csv(io.BytesIO(data), index_col=0)
        if not signaling_type is None:
            df_ligrec = df_ligrec[df_ligrec.iloc[:,3] == signaling_type]
    elif database == 'CellPhoneDB_v4.0':
        data = pkgutil.get_data(__name__, "_data/LRdatabase/CellPhoneDB_v4.0/CellPhoneDBv4.0."+species+".csv")
        df_ligrec = pd.read_csv(io.BytesIO(data), index_col=0)
        if not signaling_type is None:
            df_ligrec = df_ligrec[df_ligrec.iloc[:,3] == signaling_type]
        
    return df_ligrec

def filter_lr_database(
    df_ligrec: pd.DataFrame,
    adata: anndata.AnnData,
    heteromeric: bool = True,
    heteromeric_delimiter: str = "_",
    heteromeric_rule: str = "min",
    filter_criteria: str = "min_cell_pct",
    min_cell: int = 100,
    min_cell_pct: float = 0.05
):
    """
    Filter ligand-receptor pairs.

    Parameters
    ----------
    df_ligrec
        The pandas dataframe of ligand-receptor database with three columns being ligand, receptor, and pathway name respectively.
    adata
        The AnnData object of gene expression. Unscaled data (minimum being zero) is expected.
    heteromeric
        Whether the ligands and receptors are described as heteromeric.
    heteromeric_delimiter
        If heteromeric notations are used for ligands and receptors, the character separating the heteromeric units.
    heteromeric_rule
        When  heteromeric is True, the rule to quantify the level of a heteromeric ligand or receptor. Choose from minimum ('min') and average ('ave').
    filter_criteria
        Use either cell percentage ('min_cell_pct') or cell numbers (min_cell) to filter genes.
    min_cell
        If filter_criteria is 'min_cell', the LR-pairs with both ligand and receptor detected in greater than or equal to min_cell cells are kept.
    min_cell_pct
        If filter_criteria is 'min_cell_pct', the LR-pairs with both ligand and receptor detected in greater than or equal to min_cell_pct percentage of cells are kepts.

    Returns
    -------
    df_ligrec_filtered: pd.DataFrame
        A pandas DataFrame of the filtered ligand-receptor pairs.

    """

    data_genes = set(adata.var_names)
    ncell = adata.shape[0]
    all_genes = list(adata.var_names)
    gene_ncell = np.array( (adata.X > 0).sum(axis=0) ).reshape(-1)
    gene_mean = np.array( adata.X.mean(axis=0) )
    ligrec_list = []
    genes_keep = []
    if not heteromeric:
        tmp_genes = set(df_ligrec.iloc[:,0]).union(set(df_ligrec.iloc[:,1]))
        tmp_genes = list(tmp_genes.intersection(data_genes))
        for gene in tmp_genes:
            if not gene in all_genes: continue
            if filter_criteria == 'min_cell_prc':
                if gene_ncell[all_genes.index(gene)] / ncell >= min_cell_pct:
                    genes_keep.append(gene)
            elif filter_criteria == 'min_cell':
                if gene_ncell[all_genes.index(gene)] >= min_cell:
                    genes_keep.append(gene)
    elif heteromeric:
        tmp_genes = list(set(df_ligrec.iloc[:,0]).union(set(df_ligrec.iloc[:,1])))
        for het_gene in tmp_genes:
            genes = het_gene.split(heteromeric_delimiter)
            gene_found = True
            for gene in genes:
                if not gene in all_genes:
                    gene_found = False
            if not gene_found: continue
            keep = True
            if filter_criteria == 'min_cell_pct' and heteromeric_rule == 'min':
                for gene in genes:
                    if gene_ncell[all_genes.index(gene)] / ncell < min_cell_pct:
                        keep = False
            elif filter_criteria == 'min_cell' and heteromeric_rule == 'min':
                for gene in genes:
                    if gene_ncell[all_genes.index(gene)] < min_cell:
                        keep = False
            elif heteromeric_rule == 'ave':
                ave_ncell = []
                for gene in genes:
                    ave_ncell.append( gene_ncell[all_genes.index(gene)] )
                if filter_criteria == 'min_cell_pct':
                    if np.mean(ave_ncell) / ncell < min_cell_pct:
                        keep = False
                elif filter_criteria == 'min_cell':
                    if np.mean(ave_ncell) < min_cell:
                        keep = False
            if keep:
                genes_keep.append(het_gene)
    for i in range(df_ligrec.shape[0]):
        if df_ligrec.iloc[i,0] in genes_keep and df_ligrec.iloc[i,1] in genes_keep:
            ligrec_list.append(list(df_ligrec.iloc[i,:]))
    
    return pd.DataFrame(data=ligrec_list)
            
                
                
                



import pandas as pd
import numpy as np
from anndata import AnnData
from .helper import intersect, find_marker, extract_exp, process_anndata
from scipy.stats import pearsonr
from scipy.stats import spearmanr
from sklearn.metrics.cluster import adjusted_rand_score
import scanpy

def pearson_cal(spatrio_map: pd.DataFrame,
    ann1: AnnData,
    ann2: AnnData,
    use_marker: bool = True,
    top_n: int = 100,
    run_marker_by: str = "type",
    scale = True,
    spot_type = True
    ):
    adata1 = ann1.copy()
    adata2 = ann2.copy()
    common_genes = intersect(adata1.var.index, adata2.var.index)
    adata1 = adata1[:, common_genes]
    adata2 = adata2[:, common_genes]
    
    # adata_merge = adata1.concatenate(adata2,index_unique=None)
    # adata_merge = process_anndata(adata_merge,scale=scale,pca=False)
    
    map = pd.DataFrame(spatrio_map[['spot','cell']])
    #n1 = adata1.shape[0]
    #data1 = extract_exp(adata1).iloc[:n1]
    #data2 = extract_exp(adata2).iloc[n1:]
    data2 = extract_exp(adata2)
    data2['cell'] = data2.index
    data2 = pd.merge(map, data2, on='cell',how="left")
    grouped = data2.groupby(['spot'])
    mapped_data = grouped[data2.columns.tolist()[2:]].sum()
    tmp = scanpy.AnnData(X = mapped_data)
    
    adata_merge = adata1.concatenate(tmp,index_unique=None)
    adata_merge = process_anndata(adata_merge,scale=scale,pca=False)
    n1 = adata1.shape[0]
    data1 = extract_exp(adata_merge).iloc[:n1]
    mapped_data = extract_exp(adata_merge).iloc[n1:]

    common_genes = intersect(data1.columns, mapped_data.columns)
    data1 = data1[common_genes]
    mapped_data = mapped_data[common_genes]
    

    if use_marker:
        #sc.pp.filter_cells(adata2_copy, min_genes=200)
        #sc.pp.filter_genes(adata2_copy, min_cells=3)
        marker1 = find_marker(ann1,top_n,maker_by=run_marker_by)
        marker2 = find_marker(ann2,top_n,maker_by=run_marker_by)
        marker1.extend(marker2)
        gene_list = np.unique(marker1).tolist()
        adata1 = adata1[:, gene_list]
        adata2 = adata2[:, gene_list]
        data1 = data1[gene_list]
        mapped_data = mapped_data[gene_list]

    common_index = np.intersect1d(data1.index, mapped_data.index, assume_unique=False, return_indices=False)
    data1 = data1.loc[common_index]
    mapped_data = mapped_data.loc[common_index]
    
    if spot_type:
        map2 = pd.DataFrame(spatrio_map[['spot','spot_type']])
        spot_type_map = dict(zip(map2['spot'], map2['spot_type']))
        data1.index = data1.index.map(spot_type_map)
        mapped_data.index = mapped_data.index.map(spot_type_map)
        data1 = data1.groupby(data1.index).agg('sum')
        mapped_data = mapped_data.groupby(mapped_data.index).agg('sum')

    pearson_s = []
    for spot in data1.index.tolist():
        #x=data1.loc[spot].values
        #y=mapped_data.loc[spot].values
        from sklearn.preprocessing import scale
        x=data1.loc[spot].values
        y=mapped_data.loc[spot].values
        pc = pearsonr(x,y)
        pearson_s.append(pc[0])
    pearson = np.mean(pearson_s)
    return pearson

def calculate_cell_counts(cell_ratios: pd.DataFrame, cell_num_dict: dict) -> pd.DataFrame:
    cell_counts = cell_ratios.copy()
    for spot, cell_num in cell_num_dict.items():
        cell_counts.loc[spot] *= cell_num
    return cell_counts

def accuracy_cal(cell_counts: pd.DataFrame, ref_counts: pd.DataFrame):
    # common_spots = cell_ratios.index.intersection(ref_ratios.index)
    # cell_ratios = cell_ratios.loc[common_spots]
    # ref_ratios = ref_ratios.loc[common_spots]
    
    # if top_num is not None:
    #     cell_num = {spot: top_num for spot in common_spots}
    # if expected_num is not None:
    #     expected_num = expected_num.loc[common_spots]
    #     cell_num = expected_num['cell_num'].to_dict()
    
    # columns_order = ref_ratios.columns
    # cell_ratios = cell_ratios.reindex(columns=columns_order).fillna(0)
    # cell_ratios = cell_ratios.reindex(ref_ratios.index).fillna(0)
    # cell_counts = calculate_cell_counts(cell_ratios, cell_num).round().astype(int)
    # ref_counts = calculate_cell_counts(ref_ratios, cell_num).round().astype(int)
    common_spots = cell_counts.index.intersection(ref_counts.index)
    cell_counts = cell_counts.loc[common_spots]
    ref_counts = ref_counts.loc[common_spots]
    
    columns_order = ref_counts.columns
    cell_counts = cell_counts.reindex(columns=columns_order).fillna(0)
    cell_counts = cell_counts.reindex(ref_counts.index).fillna(0)
    
    # accuracy = []
    # for i in cell_counts.index:
    #     A = cell_counts.loc[i]
    #     B = ref_counts.loc[i]
    #     tmp = pd.DataFrame({'c1' : A,
    #                         'c2' : B})
    #     common_cells = tmp.min(axis=1).sum()
    #     total_cells = tmp.sum(axis=0).max()
    #     matching_ratio = common_cells / total_cells
    #     accuracy.append(matching_ratio)
    # accuracy = np.mean(accuracy)
    
    common_cells_sum = 0
    total_cells_sum = 0
    for i in cell_counts.index:
        A = cell_counts.loc[i]
        B = ref_counts.loc[i]
        tmp = pd.DataFrame({'c1': A, 'c2': B})
        common_cells_sum += tmp.min(axis=1).sum()
        total_cells_sum += tmp['c2'].sum()
    accuracy = common_cells_sum / total_cells_sum
    return accuracy

def spearman_cal(spatrio_map: pd.DataFrame,
    ann1: AnnData,
    ann2: AnnData,
    use_marker: bool = True,
    top_n: int = 100,
    run_marker_by: str = "type",
    scale = True,
    spot_type = True
    ):
    adata1 = ann1.copy()
    adata2 = ann2.copy()
    common_genes = intersect(adata1.var.index, adata2.var.index)
    adata1 = adata1[:, common_genes]
    adata2 = adata2[:, common_genes]
    
    # adata_merge = adata1.concatenate(adata2,index_unique=None)
    # adata_merge = process_anndata(adata_merge,scale=scale,pca=False)
    
    map = pd.DataFrame(spatrio_map[['spot','cell']])
    #n1 = adata1.shape[0]
    #data1 = extract_exp(adata1).iloc[:n1]
    #data2 = extract_exp(adata2).iloc[n1:]
    data2 = extract_exp(adata2)
    data2['cell'] = data2.index
    data2 = pd.merge(map, data2, on='cell',how="left")
    grouped = data2.groupby(['spot'])
    mapped_data = grouped[data2.columns.tolist()[2:]].sum()
    tmp = scanpy.AnnData(X = mapped_data)
    
    adata_merge = adata1.concatenate(tmp,index_unique=None)
    adata_merge = process_anndata(adata_merge,scale=scale,pca=False)
    n1 = adata1.shape[0]
    data1 = extract_exp(adata_merge).iloc[:n1]
    mapped_data = extract_exp(adata_merge).iloc[n1:]

    common_genes = intersect(data1.columns, mapped_data.columns)
    data1 = data1[common_genes]
    mapped_data = mapped_data[common_genes]
    

    if use_marker:
        #sc.pp.filter_cells(adata2_copy, min_genes=200)
        #sc.pp.filter_genes(adata2_copy, min_cells=3)
        marker1 = find_marker(ann1,top_n,maker_by=run_marker_by)
        marker2 = find_marker(ann2,top_n,maker_by=run_marker_by)
        marker1.extend(marker2)
        gene_list = np.unique(marker1).tolist()
        adata1 = adata1[:, gene_list]
        adata2 = adata2[:, gene_list]
        data1 = data1[gene_list]
        mapped_data = mapped_data[gene_list]

    common_index = np.intersect1d(data1.index, mapped_data.index, assume_unique=False, return_indices=False)
    data1 = data1.loc[common_index]
    mapped_data = mapped_data.loc[common_index]
    
    if spot_type:
        map2 = pd.DataFrame(spatrio_map[['spot','spot_type']])
        spot_type_map = dict(zip(map2['spot'], map2['spot_type']))
        data1.index = data1.index.map(spot_type_map)
        mapped_data.index = mapped_data.index.map(spot_type_map)
        data1 = data1.groupby(data1.index).agg('sum')
        mapped_data = mapped_data.groupby(mapped_data.index).agg('sum')

    spearman_s = []
    for spot in data1.index.tolist():
        x=data1.loc[spot].values
        y=mapped_data.loc[spot].values
        sc, _ = spearmanr(x,y)
        spearman_s.append(sc)
    spearman = np.mean(spearman_s)
    return spearman

def ari_cal(map_data: pd.DataFrame) :
    accuracy_data = map_data.copy()
    ari = adjusted_rand_score(accuracy_data.spot_type, accuracy_data.cell_type)
    return ari

def rmse_cal(cell_ratios, ref_ratios):
    mse = ((cell_ratios - ref_ratios) ** 2).mean(axis=1)
    rmse = np.sqrt(mse.mean())
    return rmse
